from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.dataflow as dataflow

def mkMadd():
    x = dataflow.Variable('xd', valid='xv', signed=True)
    y = dataflow.Variable('yd', valid='yv', signed=True)
    vreset = dataflow.Variable('vreset_data', valid='vreset', width=1)
    
    xy = x * y
    
    z = dataflow.Iadd(xy, signed=True, initval=0, reset=vreset)

    z.output('zd', valid='zv')
    df = dataflow.Dataflow(z)
    m = df.to_module('madd')

    return m

def mkMatmul(n=16, datawidth=32):
    m = Module('Matmul')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    # BRAM I/F
    addrwidth = int(math.log(n, 2)) * 2
    
    xaddr = m.OutputReg('xaddr', addrwidth, initval=0)
    xdin = m.Input('xdin', datawidth)
    
    yaddr = m.OutputReg('yaddr', addrwidth, initval=0)
    ydin = m.Input('ydin', datawidth)
    
    zaddr = m.OutputReg('zaddr', addrwidth, initval=0)
    zdout = m.OutputReg('zdout', datawidth, initval=0)
    zwe = m.OutputReg('zwe')

    # External Control
    start = m.Input('start')
    busy = m.OutputReg('busy', initval=0)

    # MADD Instance
    madd = mkMadd()

    ivalid = m.Reg('ivalid', initval=0)
    odata = m.Wire('odata', datawidth)
    ovalid = m.Wire('ovalid')
    vreset = m.Reg('vreset', initval=0)

    m.Instance(madd, 'madd',
               ports=[ ('CLK', clk), ('RST', rst),
                       ('xd', xdin), ('xv', ivalid), ('yd', ydin), ('yv', ivalid),
                       ('zd', odata), ('zv', ovalid),
                       ('vreset_data', vreset), ('vreset', ivalid) ])

    read_count = m.TmpReg(int(addrwidth/2)+1, initval=0)
    sum_value = m.TmpReg(datawidth, initval=0)
    sum_count = m.TmpReg(int(addrwidth/2)+1, initval=0)
    
    # main FSM
    fsm = FSM(m, 'fsm', clk, rst)

    init = fsm.current

    # initial values
    fsm.add( xaddr(0-1), yaddr(0-1), zaddr(0-1), zwe(0),
             busy(0), vreset(0), ivalid(0), 
             read_count(0), sum_value(0), sum_count(0) )

    # start
    fsm.add( busy(1), cond=start )
    fsm.goto_next(cond=start)

    comp = fsm.current

    # reset pipeline for each 1st data
    fsm.add( vreset(0), delay=1 )
    fsm.add( vreset(1), delay=1, cond=read_count==0 )
    
    # read data
    fsm.add( xaddr.inc(), yaddr.inc(), cond=read_count<n )
    fsm.add( ivalid(0), delay=1 )
    fsm.add( ivalid(1), delay=1, cond=read_count<n )
    fsm.add( read_count.inc(), cond=read_count<n )

    # comp data
    fsm.add( sum_count.inc(), cond=ovalid )
    fsm.add( sum_value(odata), cond=ovalid )
    
    # dump
    #fsm.add( Systask('display', 'xaddr=%10d yaddr=%10d', xaddr, yaddr) )
    #fsm.add( Systask('display', 'xdin =%10d ydin =%10d', xdin, ydin), delay=1 )
    #fsm.add( Systask('display', 'count=%10d odata=%10d', sum_count, odata), cond=ovalid ) 
    
    fsm.goto_next(cond=sum_count==n)

    # write data
    fsm.add( zaddr.inc() )
    fsm.add( zdout(sum_value), zwe(1) )
    fsm.add( zwe(0), delay=1 )

    # set up for next
    fsm.add( yaddr(0-1), cond=yaddr==n*n-1 )
    fsm.add( xaddr(xaddr-n), cond=yaddr<n*n-1)
    fsm.add( read_count(0), sum_count(0) )

    done = (zaddr == n * n - 2)
    fsm.goto(comp, cond=Not(done))
    fsm.goto(init, cond=done)

    fsm.make_always()

    return m

def mkTest(n=16, datawidth=32):

    m = Module('test')

    # target instance
    main = mkMatmul(n, datawidth)

    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    xaddr = ports['xaddr']
    xdin = ports['xdin']

    yaddr = ports['yaddr']
    ydin = ports['ydin']
    
    zaddr = ports['zaddr']
    zdout = ports['zdout']
    zwe = ports['zwe']

    start = ports['start']
    busy = ports['busy']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    reset_done = m.Reg('reset_done', initval=0)
    reset_stmt = []
    reset_stmt.append( reset_done(0) )
    reset_stmt.append( start(0) )
    reset_stmt.append( xdin(0) )
    reset_stmt.append( ydin(0) )

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, reset_stmt, period=100)

    nclk = simulation.next_clock

    init.add(
        Delay(1000),
        reset_done(1),
        nclk(clk),
        Delay(100000),
        Systask('finish'),
    )

    mag = 2

    fsm = FSM(m, 'fsm', clk, rst)
    
    fsm.goto_next(cond=reset_done)
    
    fsm.goto_next(cond=Not(busy))
    
    fsm.add( start(1) )
    fsm.add( start(0), delay=1 )
    #fsm.add( xdin(Mux(xaddr % n == xaddr / n, 1, 0)) )
    fsm.add( xdin(xaddr) )
    fsm.add( ydin(Mux(yaddr % n == yaddr / n, mag, 0)) )
    fsm.goto_next()
    
    #fsm.add( xdin(Mux(xaddr % n == xaddr / n, 1, 0)) )
    fsm.add( xdin(xaddr) )
    fsm.add( ydin(Mux(yaddr % n == yaddr / n, mag, 0)) )
    fsm.goto_next(cond=busy)

    #fsm.add( xdin(Mux(xaddr % n == xaddr / n, 1, 0)) )
    fsm.add( xdin(xaddr) )
    fsm.add( ydin(Mux(yaddr % n == yaddr / n, mag, 0)) )
    fsm.goto_next(cond=Not(busy))

    fsm.add( Systask('finish') )

    fsm.make_always()

    m.Always(Posedge(clk))(
        If(zwe)(
            Systask('display', 'zaddr=%10d zdout=%10d', zaddr, zdout),
            #If(AndList(zaddr % n == zaddr / n, zdout != 1))(
            #    Systask('display', '## wrong')
            #),
            #If(AndList(zaddr % n != zaddr / n, zdout != 0))(
            #    Systask('display', '## wrong')
            #)
            If(zaddr * mag != zdout)(
                Systask('display', '## wrong result')
            ),
        )
    )
    
    return m

if __name__ == '__main__':
    n = 16
    test = mkTest(n)
    verilog = test.to_verilog('tmp.v')
    #print(verilog)

    # run simulator (Icarus Verilog)
    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # only target RTL
    #main = mkMatmul(n)
    #verilog = main.to_verilog('tmp.v')
    #print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import _while

expected_verilog = """
module test;
  reg CLK;
  reg RST; 
  reg [32-1:0] count;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, CLK, RST, count);
  end

  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end

  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;

    count = 0;
    while(count < 1024) begin
      count = count + 1;
      @(posedge CLK);
    end

    $finish;
  end
endmodule
"""

def test():
    veriloggen.reset()
    test_module = _while.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *

def mkBram(datawidth=32, addrwidth=10, numports=2):
    m = Module('BRAM%d' % numports)
    
    clk = m.Input('CLK')
    ports = []
    for i in range(numports):
        addr = m.Input('ADDR%d' % i, addrwidth)
        din = m.Input('DIN%d' % i, datawidth)
        we = m.Input('WE%d' % i)
        dout = m.Output('DOUT%d' % i, datawidth)
        delay_addr = m.Reg('delay_ADDR%d' % i, addrwidth)
        ports.append( (addr, din, we, dout, delay_addr) )

    mem = m.Reg('mem', datawidth, length=2**addrwidth)

    for i in range(numports):
        addr, din ,we, dout, delay_addr = ports[i]
        m.Always(Posedge(clk))(
            If(we)(
                mem[addr](din)
            ),
            delay_addr(addr)
        )
        m.Assign(dout(mem[delay_addr]))

    return m

def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2

    bram = mkBram(datawidth, addrwidth, numports)

    bram_ports = []
    for i in range(numports):
        addr = m.Reg('bram_addr%d' % i, addrwidth)
        din = m.Reg('bram_din%d' % i, datawidth)
        we = m.Reg('bram_we%d' % i)
        dout = m.Wire('bram_dout%d' % i, datawidth)
        bram_ports.append( [addr, din, we, dout] )

    ports = [ clk ]
    for bram_port in bram_ports:
        ports.extend(bram_port)

    # BRAM instance
    m.Instance(bram, 'bram_inst', params=(), ports=ports)

    # example how to access BRAM
    fsm = FSM(m, 'fsm', clk, rst)
    
    for addr, din, we, dout in bram_ports:
        fsm.add( addr(0), din(0), we(0) )

    fsm.goto_next()

    count = m.Reg('count', 32, initval=0)
    for port, (addr, din, we, dout) in enumerate(bram_ports):
        fsm.add( addr(count+port*n//numports), din(count), we(1), count.inc() )

    fsm.goto_next(cond=count==n//numports-1)
    
    fsm.make_always()
    
    return m
    
def mkTest():
    m = Module('test')
    
    # target instance
    main = mkMain()
    
    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)
    
    clk = ports['CLK']
    rst = ports['RST']
    
    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))
    
    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)
    
    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))))

from veriloggen import *
import veriloggen.types.axi as axi


def mkMain():
    m = Module('main')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    myaxi = axi.AxiSlave(m, 'myaxi', clk, rst)
    myaxi.disable_write()

    fsm = FSM(m, 'fsm', clk, rst)

    # read address
    addr, counter, valid = myaxi.pull_read_request(cond=fsm)
    fsm.If(valid).goto_next()

    # read rdata
    rdata = m.Reg('rdata', 32, initval=512)
    fsm.goto_next()

    ack, last = myaxi.push_read_data(rdata, counter, cond=fsm)
    fsm.If(ack)(
        rdata(rdata + 512)
    )
    fsm.If(last).goto_next()

    fsm.If(rdata < 4096).goto_init()
    fsm.If(rdata >= 4096).goto_next()

    seq = Seq(m, 'seq', clk, rst)
    sum = m.Reg('sum', 32, initval=0)
    expected_sum = (512 + 512 * 7) * 7 // 2
    
    seq.If(myaxi.rdata.rvalid, myaxi.rdata.rready)(
        sum(sum + myaxi.rdata.rdata)
    )
    
    fsm(
        Systask('display', "sum=%d expected_sum=%d", sum, expected_sum)
    )
    fsm.goto_next()
    
    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    awvalid = ports['myaxi_awvalid']
    _awvalid = m.TmpWireLike(awvalid)
    _awvalid.assign(0)
    m.Always()(awvalid(_awvalid))

    awaddr = ports['myaxi_awaddr']
    _awaddr = m.TmpWireLike(awaddr)
    _awaddr.assign(0)
    m.Always()(awaddr(_awaddr))

    awlen = ports['myaxi_awlen']
    _awlen = m.TmpWireLike(awlen)
    _awlen.assign(0)
    m.Always()(awlen(_awlen))

    arvalid = ports['myaxi_arvalid']
    _arvalid = m.TmpWireLike(arvalid)
    _arvalid.assign(1)  # on
    m.Always()(arvalid(_arvalid))

    araddr = ports['myaxi_araddr']
    _araddr = m.TmpWireLike(araddr)
    _araddr.assign(0x100)  # on
    m.Always()(araddr(_araddr))

    arlen = ports['myaxi_arlen']
    _arlen = m.TmpWireLike(arlen)
    _arlen.assign(0)  # on
    m.Always()(arlen(_arlen))

    wvalid = ports['myaxi_wvalid']
    _wvalid = m.TmpWireLike(wvalid)
    _wvalid.assign(0)
    m.Always()(wvalid(_wvalid))

    wdata = ports['myaxi_wdata']
    _wdata = m.TmpWireLike(wdata)
    _wdata.assign(0)
    m.Always()(wdata(_wdata))

    wstrb = ports['myaxi_wstrb']
    _wstrb = m.TmpWireLike(wstrb)
    _wstrb.assign(Repeat(Int(1, 1), 32 // 8))
    m.Always()(wstrb(_wstrb))

    wlast = ports['myaxi_wlast']
    _wlast = m.TmpWireLike(wlast)
    _wlast.assign(0)
    m.Always()(wlast(_wlast))

    rready = ports['myaxi_rready']
    _rready = m.TmpWireLike(rready)
    _rready.assign(1)  # on
    m.Always()(rready(_rready))

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module


def mkMultiplierCore(index, lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    retwidth = lwidth + rwidth

    m = module.Module('multiplier_core_%d' % index)

    clk = m.Input('CLK')
    update = m.Input('update')

    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)

    _a = m.Reg('_a', lwidth, signed=lsigned)
    _b = m.Reg('_b', rwidth, signed=rsigned)
    tmpval = [m.Reg('_tmpval%d' % i, retwidth, signed=True)
              for i in range(depth - 1)]
    rslt = m.Wire('rslt', retwidth, signed=True)

    __a = _a
    __b = _b
    if not lsigned:
        __a = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _a))
    if not rsigned:
        __b = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _b))

    m.Assign(rslt(__a * __b))
    m.Assign(c(tmpval[depth - 2]))

    m.Always(vtypes.Posedge(clk))(
        vtypes.If(update)(
            _a(a),
            _b(b),
            tmpval[0](rslt),
            [tmpval[i](tmpval[i - 1]) for i in range(1, depth - 1)]
        ))

    return m


def mkMultiplier(index, lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    if lwidth < 0:
        raise ValueError("data width must be greater than 0.")
    if rwidth < 0:
        raise ValueError("data width must be greater than 0.")
    if depth < 2:
        raise ValueError("depth must be greater than 2.")

    retwidth = lwidth + rwidth

    mult = mkMultiplierCore(index, lwidth, rwidth, lsigned, rsigned, depth)

    m = module.Module('multiplier_%d' % index)

    clk = m.Input('CLK')
    rst = m.Input('RST')

    update = m.Input('update')
    enable = m.Input('enable')
    valid = m.Output('valid')

    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)

    valid_reg = [m.Reg('valid_reg%d' % i) for i in range(depth)]

    m.Assign(valid(valid_reg[depth - 1]))

    m.Always(vtypes.Posedge(clk))(
        vtypes.If(rst)(
            [valid_reg[i](0) for i in range(depth)]
        ).Else(
            vtypes.If(update)(
                valid_reg[0](enable),
                [valid_reg[i](valid_reg[i - 1]) for i in range(1, depth)]
            )
        ))

    ports = [('CLK', clk), ('update', update), ('a', a), ('b', b), ('c', c)]
    m.Instance(mult, 'mult', ports=ports)

    return m

# global multiplier count
index_count = 0


def get_mul(lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    global index_count
    mul = mkMultiplier(index_count, lwidth, rwidth, lsigned, rsigned, depth)
    index_count += 1
    return mul


def reset():
    global index_count
    index_count = 0



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.dataflow as dataflow


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    df = dataflow.DataflowManager(m, clk, rst)
    #df.enable_draw_graph()

    # input with register
    xdata = m.Reg('xdata', 32, initval=0)
    xvalid = m.Reg('xvalid', initval=0)
    xready = m.Wire('xready')
    x = df.Variable(xdata, xvalid, xready)

    # input with name
    y = df.Variable('ydata', 'yvalid', 'yready')

    # output
    a = x + y + 1
    b = a + x + y

    a.output('adata', 'avalid', 'aready')
    b.output('bdata', 'bvalid', 'bready')

    # write
    xfsm = FSM(m, 'xfsm', clk, rst)
    xcount = m.TmpReg(32, initval=0)

    xack = x.write(xcount, cond=xfsm)
    xfsm.If(xack)(
        xcount.inc()
    )
    xfsm.Then().If(xcount == 15).goto_next()

    # write
    yfsm = FSM(m, 'yfsm', clk, rst)
    ycount = m.TmpReg(32, initval=0)

    yack = y.write(ycount, cond=yfsm)
    yfsm.If(yack)(
        ycount(ycount + 1)
    )
    yfsm.Then().If(ycount == 15).goto_next()

    # read
    aseq = Seq(m, 'aseq', clk, rst)

    adata, avalid = a.read(cond=aseq)
    aseq.If(avalid)(
        Systask('display', "adata=%d", adata)
    )

    # read
    bseq = Seq(m, 'bseq', clk, rst)

    bdata, bvalid = b.read(cond=bseq)
    bseq.If(bvalid)(
        Systask('display', "bdata=%d", bdata)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import chatter_clear

expected_verilog = """
module test #
(
  parameter length = 1024
);

  reg CLK;
  reg RST;
  reg din;
  wire dout;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _tmp_0;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _tmp_0 <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          din <= 0;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 2000) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 2000) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          din <= 1;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 10) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 10) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          din <= 0;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 10) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 10) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          din <= 1;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 2000) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 2000) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          din <= 0;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 10) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 10) begin
            fsm <= fsm_5;
          end 
        end
        fsm_5: begin
          din <= 1;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 10) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 10) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          din <= 0;
          _tmp_0 <= _tmp_0 + 1;
          if(_tmp_0 == 2000) begin
            _tmp_0 <= 0;
          end 
          if(_tmp_0 == 2000) begin
            fsm <= fsm_7;
          end 
        end
      endcase
    end
  end


  chatter_clear
  #(
    .length(length)
  )
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .din(din),
    .dout(dout)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, din, dout, fsm, _tmp_0);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    fsm = fsm_init;
    _tmp_0 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module chatter_clear #
(
  parameter length = 1024
)
(
  input CLK,
  input RST,
  input din,
  output reg dout
);

  reg [32-1:0] _tmp_0;

  always @(posedge CLK) begin
    if(RST) begin
      dout <= 0;
    end else begin
      if(din == dout) begin
        _tmp_0 <= 0;
      end 
      if(din != dout) begin
        _tmp_0 <= _tmp_0 + 1;
      end 
      if(_tmp_0 == length) begin
        _tmp_0 <= 0;
      end 
      if(_tmp_0 == length) begin
        dout <= din;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = chatter_clear.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))))

from veriloggen import *
import veriloggen.types.ram as ram
import veriloggen.dataflow as dataflow


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2
    myram = ram.SyncRAMManager(m, 'myram', clk, rst, datawidth, addrwidth, 2, nodataflow=True)

    xfsm = FSM(m, 'xfsm', clk, rst)
    xaddr = m.Reg('xaddr', 32, initval=0)
    xcount = m.Reg('xcount', 32, initval=0)

    # dataflow variables without manager
    x = dataflow.Variable('xdata', 'xvalid', 'xready')
    y = x + 100
    y.output('ydata', 'yvalid', 'yread')

    df = dataflow.Dataflow(y)
    df.implement(m, clk, rst)

    # Initialization
    xfsm(
        xaddr(0),
        xcount(0),
    )
    xfsm.goto_next()

    # write data to RAM
    step = 16
    myram.write(0, xaddr, xcount, cond=xfsm)
    xfsm(
        xaddr.inc(),
        xcount.inc()
    )
    xfsm.If(xcount == step - 1)(
        xaddr(0),
        xcount(0)
    )
    xfsm.Then().goto_next()

    # read data from RAM
    read_data, read_valid = myram.read(0, xaddr, cond=(xfsm, xaddr < step))
    # RAM -> dataflow
    xack = x.write(read_data, cond=read_valid)
    xfsm(
        xaddr.inc()
    )
    xfsm.If(read_valid)(
        Systask('display', 'RAM0[%d] = %d', xfsm.Prev(xaddr, 2), read_data)
    )
    xfsm.If(xfsm.Prev(xaddr, 1) == step).goto_next()

    # write result to RAM
    yfsm = FSM(m, 'yfsm', clk, rst)
    yaddr = m.Reg('yaddr', 32, initval=0)
    yfsm(
        yaddr(0)
    )
    yfsm.goto_next()

    # read from dataflow
    rdata, rvalid = y.read(yfsm)
    # dataflow -> RAM
    myram.write(1, yaddr, rdata, cond=(yfsm, rvalid))
    yfsm.If(rvalid)(
        yaddr.inc()
    )
    yfsm.If(yaddr == step - 1)(
        yaddr(0)
    )
    yfsm.Then().goto_next()

    # read data from RAM
    read_data, read_valid = myram.read(1, yaddr, cond=(yfsm, yaddr < step))
    yfsm.If(yaddr < step)(
        yaddr.inc()
    )
    yfsm.If(read_valid)(
        Systask('display', 'RAM1[%d] = %d', yfsm.Prev(yaddr, 2), read_data)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from veriloggen import *

def mkUserlogic():
    m = Module('userlogic')
    data_width = m.Parameter('DATA_WIDTH', 32)
    addr_len = m.Parameter('ADDR_LEN', 10)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    
    mem_addr = m.OutputReg('mem_addr', addr_len, initval=0)
    mem_wdata = m.OutputReg('mem_wdata', data_width, initval=0)
    mem_rdata = m.Input('mem_rdata', data_width)
    mem_wvalid = m.OutputReg('mem_wvalid', initval=0)
    mem_rvalid = m.OutputReg('mem_rvalid', initval=0)
    ch_wdata = m.OutputReg('ch_wdata', data_width, initval=0)
    ch_enq = m.OutputReg('ch_enq', initval=0)
    ch_almfull = m.Input('ch_almfull')
    ch_rdata = m.Input('ch_rdata', data_width)
    ch_deq = m.OutputReg('ch_deq', initval=0)
    ch_empty = m.Input('ch_empty')

    # Finite State Machine
    fsm = FSM(m, 'fsm', clk, rst)

    read_count = m.Reg('read_count', width=32, initval=0)
    sum = m.Reg('sum', width=32, initval=0)
    size = m.Reg('size', width=32, initval=0)

    # building FSM
    fsm.goto_next()
    fsm.goto_next()
    start = fsm.current

    fsm.add( ch_deq(1), cond=Not(ch_empty) )
    fsm.add( ch_deq(0), delay=1 )
    fsm.add( size(ch_rdata), delay=2 )
    fsm.goto_next(cond=Not(ch_empty))
    
    fsm.add( mem_addr( -Int(1) ))
    fsm.goto_next()
    
    fsm.add( mem_rvalid(1), mem_addr.inc(), read_count.inc() )
    fsm.add( mem_rvalid(0), delay=1 )
    fsm.add( sum.add(mem_rdata), delay=2, keep=2 )
    fsm.goto_next( cond=(read_count==size-1) )
    fsm.goto_next()
    
    fsm.add( Systask('display', 'sum=%d', sum), delay=2)
    fsm.add( ch_enq(1), cond=Not(ch_almfull) )
    fsm.add( ch_enq(0), delay=1 )
    fsm.add( ch_wdata(sum) )
    fsm.goto(start)
    
    # building always statement
    fsm.make_always(reset=m.make_reset())
    
    return m

def mkTest():
    m = Module('test')
    data_width = m.Parameter('DATA_WIDTH', 32)
    addr_len = m.Parameter('ADDR_LEN', 10)
    clk = m.Reg('CLK')
    rst = m.Reg('RST')
    mem_addr = m.Wire('mem_addr', addr_len)
    mem_wdata = m.Wire('mem_wdata', data_width)
    mem_rdata = m.Reg('mem_rdata', data_width, initval=0)
    mem_wvalid = m.Wire('mem_wvalid')
    mem_rvalid = m.Wire('mem_rvalid')
    ch_wdata = m.Wire('ch_wdata', data_width)
    ch_enq = m.Wire('ch_enq')
    ch_almfull = m.Reg('ch_almfull', initval=0)
    ch_rdata = m.Reg('ch_rdata', data_width, initval=0)
    ch_deq = m.Wire('ch_deq')
    ch_empty = m.Reg('ch_empty', initval=1)

    uut = m.Instance(mkUserlogic(), 'uut',
                     params=connect_same_name(data_width, addr_len),
                     ports=connect_same_name(clk, rst,
                                             mem_addr, mem_wdata, mem_rdata,
                                             mem_wvalid, mem_rvalid,
                                             ch_wdata, ch_enq, ch_almfull,
                                             ch_rdata, ch_deq, ch_empty))

    fsm = FSM(m, 'test_fsm', clk, rst)
    simulation.setup_waveform(m, uut, fsm.state)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    #init.add(
    #    Delay(1000),
    #    Systask('finish'),
    #)

    for i in range(10):
        fsm.goto_next()
    
    fsm.add( ch_empty(0) )
    fsm.add( ch_rdata(10) )
    fsm.goto_next()
    
    fsm.add( ch_empty(1), cond=ch_deq)
    fsm.add( mem_rdata(0) )
    fsm.goto_next(cond=ch_deq)

    fsm.add( mem_rdata.inc(), cond=mem_rvalid, delay=1 )
    fsm.goto_next(cond=ch_enq)

    fsm.add( Systask('finish') )
    
    fsm.make_always()
    
    return m

if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    filename = os.path.dirname(os.path.abspath(__file__)) + '/led.v'
    modules = from_verilog.read_verilog_module(filename)
    m = modules['blinkled']
    return m

def mkTop():
    m = Module('top')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', 8)
    
    params = ( )
    ports = ( clk, rst, led )

    led = mkLed()
    m.Instance(led, 'inst_blinkled', params, ports)

    return m

if __name__ == '__main__':
    top = mkTop()
    verilog = top.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import seq_merge

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST
);

  reg [32-1:0] x;
  reg [32-1:0] y;
  reg [32-1:0] z;
  reg _seq_x_cond_0_1;
  reg _seq_y_cond_0_1;
  reg _seq_y_cond_0_2;
  reg [32-1:0] _x_1;
  reg [32-1:0] _y_1;

  always @(posedge CLK) begin
    if(RST) begin
      x <= 0;
      _seq_x_cond_0_1 <= 0;
      _x_1 <= 0;
      y <= 0;
      _seq_y_cond_0_1 <= 0;
      _seq_y_cond_0_2 <= 0;
      _y_1 <= 0;
      z <= 0;
    end else begin
      if(_seq_y_cond_0_2) begin
        y <= 0;
      end 
      if(_seq_x_cond_0_1) begin
        x <= 0;
      end 
      _seq_y_cond_0_2 <= _seq_y_cond_0_1;
      if(x < 16) begin
        x <= x + 1;
      end 
      _seq_x_cond_0_1 <= x == 16;
      _x_1 <= x;
      if(y < 32) begin
        y <= y + 1;
      end 
      _seq_y_cond_0_1 <= y == 32;
      _y_1 <= y;
      z <= _x_1 + _y_1;
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = seq_merge.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_manager_readwrite

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [32-1:0] xdata;
  reg xvalid;
  wire xready;
  reg [32-1:0] xfsm;
  localparam xfsm_init = 0;
  reg [32-1:0] _tmp_0;
  reg __dataflow_seq_0_cond_0_1;
  reg [32-1:0] yfsm;
  localparam yfsm_init = 0;
  reg [32-1:0] _tmp_1;
  wire [32-1:0] ydata;
  wire yvalid;
  wire yready;
  reg [32-1:0] _tmp_2;
  assign ydata = _tmp_2;
  reg _tmp_3;
  assign yvalid = _tmp_3;
  reg __dataflow_seq_0_cond_1_1;
  wire [32-1:0] adata;
  wire avalid;
  wire aready;
  assign aready = 1;
  wire [32-1:0] bdata;
  wire bvalid;
  wire bready;
  assign bready = 1;
  reg [32-1:0] _tmp_data_4;
  reg _tmp_valid_4;
  wire _tmp_ready_4;
  assign xready = (_tmp_ready_4 || !_tmp_valid_4) && (xvalid && yvalid) && ((_tmp_ready_5 || !_tmp_valid_5) && xvalid);
  assign yready = (_tmp_ready_4 || !_tmp_valid_4) && (xvalid && yvalid) && ((_tmp_ready_6 || !_tmp_valid_6) && yvalid);
  assign _tmp_ready_4 = (_tmp_ready_7 || !_tmp_valid_7) && _tmp_valid_4;
  reg [32-1:0] _tmp_data_5;
  reg _tmp_valid_5;
  wire _tmp_ready_5;
  assign _tmp_ready_5 = (_tmp_ready_8 || !_tmp_valid_8) && _tmp_valid_5;
  reg [32-1:0] _tmp_data_6;
  reg _tmp_valid_6;
  wire _tmp_ready_6;
  assign _tmp_ready_6 = (_tmp_ready_9 || !_tmp_valid_9) && _tmp_valid_6;
  reg [32-1:0] _tmp_data_7;
  reg _tmp_valid_7;
  wire _tmp_ready_7;
  assign _tmp_ready_7 = (_tmp_ready_10 || !_tmp_valid_10) && (_tmp_valid_7 && _tmp_valid_8) && ((_tmp_ready_12 || !_tmp_valid_12) && _tmp_valid_7);
  reg [32-1:0] _tmp_data_8;
  reg _tmp_valid_8;
  wire _tmp_ready_8;
  assign _tmp_ready_8 = (_tmp_ready_10 || !_tmp_valid_10) && (_tmp_valid_7 && _tmp_valid_8);
  reg [32-1:0] _tmp_data_9;
  reg _tmp_valid_9;
  wire _tmp_ready_9;
  assign _tmp_ready_9 = (_tmp_ready_11 || !_tmp_valid_11) && _tmp_valid_9;
  reg [32-1:0] _tmp_data_10;
  reg _tmp_valid_10;
  wire _tmp_ready_10;
  assign _tmp_ready_10 = (_tmp_ready_13 || !_tmp_valid_13) && (_tmp_valid_10 && _tmp_valid_11);
  reg [32-1:0] _tmp_data_11;
  reg _tmp_valid_11;
  wire _tmp_ready_11;
  assign _tmp_ready_11 = (_tmp_ready_13 || !_tmp_valid_13) && (_tmp_valid_10 && _tmp_valid_11);
  reg [32-1:0] _tmp_data_12;
  reg _tmp_valid_12;
  wire _tmp_ready_12;
  assign _tmp_ready_12 = (_tmp_ready_14 || !_tmp_valid_14) && _tmp_valid_12;
  reg [32-1:0] _tmp_data_13;
  reg _tmp_valid_13;
  wire _tmp_ready_13;
  reg [32-1:0] _tmp_data_14;
  reg _tmp_valid_14;
  wire _tmp_ready_14;
  assign bdata = _tmp_data_13;
  assign bvalid = _tmp_valid_13;
  assign _tmp_ready_13 = bready;
  assign adata = _tmp_data_14;
  assign avalid = _tmp_valid_14;
  assign _tmp_ready_14 = aready;

  always @(posedge CLK) begin
    if(RST) begin
      xdata <= 0;
      xvalid <= 0;
      __dataflow_seq_0_cond_0_1 <= 0;
      _tmp_2 <= 0;
      _tmp_3 <= 0;
      __dataflow_seq_0_cond_1_1 <= 0;
      _tmp_data_4 <= 0;
      _tmp_valid_4 <= 0;
      _tmp_data_5 <= 0;
      _tmp_valid_5 <= 0;
      _tmp_data_6 <= 0;
      _tmp_valid_6 <= 0;
      _tmp_data_7 <= 0;
      _tmp_valid_7 <= 0;
      _tmp_data_8 <= 0;
      _tmp_valid_8 <= 0;
      _tmp_data_9 <= 0;
      _tmp_valid_9 <= 0;
      _tmp_data_10 <= 0;
      _tmp_valid_10 <= 0;
      _tmp_data_11 <= 0;
      _tmp_valid_11 <= 0;
      _tmp_data_12 <= 0;
      _tmp_valid_12 <= 0;
      _tmp_data_13 <= 0;
      _tmp_valid_13 <= 0;
      _tmp_data_14 <= 0;
      _tmp_valid_14 <= 0;
    end else begin
      if(__dataflow_seq_0_cond_0_1) begin
        xvalid <= 0;
      end 
      if(__dataflow_seq_0_cond_1_1) begin
        _tmp_3 <= 0;
      end 
      if((xfsm == 0) && (xready || !xvalid)) begin
        xdata <= _tmp_0;
      end 
      if((xfsm == 0) && (xready || !xvalid)) begin
        xvalid <= 1;
      end 
      __dataflow_seq_0_cond_0_1 <= 1;
      if(xvalid && !xready) begin
        xvalid <= xvalid;
      end 
      if((yfsm == 0) && (yready || !_tmp_3)) begin
        _tmp_2 <= _tmp_1;
      end 
      if((yfsm == 0) && (yready || !_tmp_3)) begin
        _tmp_3 <= 1;
      end 
      __dataflow_seq_0_cond_1_1 <= 1;
      if(_tmp_3 && !yready) begin
        _tmp_3 <= _tmp_3;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && (xready && yready) && (xvalid && yvalid)) begin
        _tmp_data_4 <= xdata + ydata;
      end 
      if(_tmp_valid_4 && _tmp_ready_4) begin
        _tmp_valid_4 <= 0;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && (xready && yready)) begin
        _tmp_valid_4 <= xvalid && yvalid;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && xready && xvalid) begin
        _tmp_data_5 <= xdata;
      end 
      if(_tmp_valid_5 && _tmp_ready_5) begin
        _tmp_valid_5 <= 0;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && xready) begin
        _tmp_valid_5 <= xvalid;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && yready && yvalid) begin
        _tmp_data_6 <= ydata;
      end 
      if(_tmp_valid_6 && _tmp_ready_6) begin
        _tmp_valid_6 <= 0;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && yready) begin
        _tmp_valid_6 <= yvalid;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_4 && _tmp_valid_4) begin
        _tmp_data_7 <= _tmp_data_4 + 2'd1;
      end 
      if(_tmp_valid_7 && _tmp_ready_7) begin
        _tmp_valid_7 <= 0;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_4) begin
        _tmp_valid_7 <= _tmp_valid_4;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && _tmp_ready_5 && _tmp_valid_5) begin
        _tmp_data_8 <= _tmp_data_5;
      end 
      if(_tmp_valid_8 && _tmp_ready_8) begin
        _tmp_valid_8 <= 0;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && _tmp_ready_5) begin
        _tmp_valid_8 <= _tmp_valid_5;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && _tmp_ready_6 && _tmp_valid_6) begin
        _tmp_data_9 <= _tmp_data_6;
      end 
      if(_tmp_valid_9 && _tmp_ready_9) begin
        _tmp_valid_9 <= 0;
      end 
      if((_tmp_ready_9 || !_tmp_valid_9) && _tmp_ready_6) begin
        _tmp_valid_9 <= _tmp_valid_6;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && (_tmp_ready_7 && _tmp_ready_8) && (_tmp_valid_7 && _tmp_valid_8)) begin
        _tmp_data_10 <= _tmp_data_7 + _tmp_data_8;
      end 
      if(_tmp_valid_10 && _tmp_ready_10) begin
        _tmp_valid_10 <= 0;
      end 
      if((_tmp_ready_10 || !_tmp_valid_10) && (_tmp_ready_7 && _tmp_ready_8)) begin
        _tmp_valid_10 <= _tmp_valid_7 && _tmp_valid_8;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && _tmp_ready_9 && _tmp_valid_9) begin
        _tmp_data_11 <= _tmp_data_9;
      end 
      if(_tmp_valid_11 && _tmp_ready_11) begin
        _tmp_valid_11 <= 0;
      end 
      if((_tmp_ready_11 || !_tmp_valid_11) && _tmp_ready_9) begin
        _tmp_valid_11 <= _tmp_valid_9;
      end 
      if((_tmp_ready_12 || !_tmp_valid_12) && _tmp_ready_7 && _tmp_valid_7) begin
        _tmp_data_12 <= _tmp_data_7;
      end 
      if(_tmp_valid_12 && _tmp_ready_12) begin
        _tmp_valid_12 <= 0;
      end 
      if((_tmp_ready_12 || !_tmp_valid_12) && _tmp_ready_7) begin
        _tmp_valid_12 <= _tmp_valid_7;
      end 
      if((_tmp_ready_13 || !_tmp_valid_13) && (_tmp_ready_10 && _tmp_ready_11) && (_tmp_valid_10 && _tmp_valid_11)) begin
        _tmp_data_13 <= _tmp_data_10 + _tmp_data_11;
      end 
      if(_tmp_valid_13 && _tmp_ready_13) begin
        _tmp_valid_13 <= 0;
      end 
      if((_tmp_ready_13 || !_tmp_valid_13) && (_tmp_ready_10 && _tmp_ready_11)) begin
        _tmp_valid_13 <= _tmp_valid_10 && _tmp_valid_11;
      end 
      if((_tmp_ready_14 || !_tmp_valid_14) && _tmp_ready_12 && _tmp_valid_12) begin
        _tmp_data_14 <= _tmp_data_12;
      end 
      if(_tmp_valid_14 && _tmp_ready_14) begin
        _tmp_valid_14 <= 0;
      end 
      if((_tmp_ready_14 || !_tmp_valid_14) && _tmp_ready_12) begin
        _tmp_valid_14 <= _tmp_valid_12;
      end 
    end
  end

  localparam xfsm_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      xfsm <= xfsm_init;
      _tmp_0 <= 0;
    end else begin
      case(xfsm)
        xfsm_init: begin
          if(xready || !xvalid) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if((xready || !xvalid) && (_tmp_0 == 15)) begin
            xfsm <= xfsm_1;
          end 
        end
      endcase
    end
  end

  localparam yfsm_1 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      yfsm <= yfsm_init;
      _tmp_1 <= 0;
    end else begin
      case(yfsm)
        yfsm_init: begin
          if(yready || !_tmp_3) begin
            _tmp_1 <= _tmp_1 + 1;
          end 
          if((yready || !_tmp_3) && (_tmp_1 == 15)) begin
            yfsm <= yfsm_1;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(avalid) begin
      $display("adata=%d", adata);
    end 
  end


  always @(posedge CLK) begin
    if(bvalid) begin
      $display("bdata=%d", bdata);
    end 
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = dataflow_manager_readwrite.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = Module('blinkled')
    clk = m.Input('CLK')
    rst = m.Input('RST')
    valid = m.OutputReg('valid', initval=0)

    fsm = FSM(m, 'fsm', clk, rst)
    
    for i in range(2):
        fsm.goto_next()

    # assert valid, then de-assert at the next cycle
    fsm(
        valid(1)
    )
    fsm.Delay(1)(
        valid(0)
    )
    
    for i in range(4):
        fsm.goto_next()

    # assert valid, then de-assert at the next cycle if not asserted again
    for i in range(4):
        fsm(
            valid(1)
        )
        fsm.Delay(1)(
            valid(0)
        )
        fsm.goto_next()
    
    fsm.make_always()

    return m

def mkTest():
    m = Module('test')
    clk = m.Reg('CLK')
    rst = m.Reg('RST')
    valid = m.Wire('valid')

    uut = m.Instance(mkLed(), 'uut',
                     ports=(('CLK', clk), ('RST', rst), ('valid', valid)))

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, period=100)

    init.add(
        Delay(1000),
        Systask('finish'),
    )

    return m
    
if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import resolver_nested_module

expected_verilog = """
module blinkled #
(
  parameter TOP_WIDTH = 16,
  parameter TOP_INC = 1
)
(
  input CLK,
  input RST,
  output [16-1:0] LED0,
  output [16-1:0] LED1
);


  sub_blinkled
  #(
    .WIDTH(16),
    .INC(2)
  )
  inst_sub_blinkled_0
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED0)
  );


  sub_blinkled_copy_0
  #(
    .WIDTH(16),
    .INC(3)
  )
  inst_sub_blinkled_1
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED1)
  );


endmodule



module sub_blinkled #
(
  parameter WIDTH = 8,
  parameter INC = 1
)
(
  input CLK,
  input RST,
  output reg [16-1:0] LED
);

  reg [26-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 2;
      end
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      if(count == 1023) begin
        LED <= LED + 2;
      end 
    end
  end


endmodule



module sub_blinkled_copy_0 #
(
  parameter WIDTH = 8,
  parameter INC = 1
)
(
  input CLK,
  input RST,
  output reg [16-1:0] LED
);

  reg [26-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 3;
      end
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      if(count == 1023) begin
        LED <= LED + 3;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = resolver_nested_module.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import resolver_for_loop

expected_verilog = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg [32-1:0] count [0:8-1];
  integer i;

  always @(posedge CLK) begin
    if(RST) begin
      count[0] <= 0;
      count[1] <= 0;
      count[2] <= 0;
      count[3] <= 0;
      count[4] <= 0;
      count[5] <= 0;
      count[6] <= 0;
      count[7] <= 0;
    end else begin
      if(count[0] == 1023) begin
        count[0] <= 0;
      end else begin
        count[0] <= count[0] + 0;
      end
      if(count[1] == 1023) begin
        count[1] <= 0;
      end else begin
        count[1] <= count[1] + 1;
      end
      if(count[2] == 1023) begin
        count[2] <= 0;
      end else begin
        count[2] <= count[2] + 2;
      end
      if(count[3] == 1023) begin
        count[3] <= 0;
      end else begin
        count[3] <= count[3] + 3;
      end
      if(count[4] == 1023) begin
        count[4] <= 0;
      end else begin
        count[4] <= count[4] + 4;
      end
      if(count[5] == 1023) begin
        count[5] <= 0;
      end else begin
        count[5] <= count[5] + 5;
      end
      if(count[6] == 1023) begin
        count[6] <= 0;
      end else begin
        count[6] <= count[6] + 6;
      end
      if(count[7] == 1023) begin
        count[7] <= 0;
      end else begin
        count[7] <= count[7] + 7;
      end
    end
  end

  reg [32-1:0] _sum;

  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      _sum = 0;
      _sum = _sum + count[0];
      _sum = _sum + count[1];
      _sum = _sum + count[2];
      _sum = _sum + count[3];
      _sum = _sum + count[4];
      _sum = _sum + count[5];
      _sum = _sum + count[6];
      _sum = _sum + count[7];
      if(_sum == 8184) begin
        LED <= LED + 1;
      end 
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = resolver_for_loop.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import stub_withcode

expected_verilog = """
module top #
  (
   parameter WIDTH = 8
  )
  (
   input CLK, 
   input RST, 
   output [WIDTH-1:0] LED
  );
  blinkled #
  (
   WIDTH
  )
  inst_blinkled
  (
   CLK,
   RST,
   LED
  );
endmodule

module blinkled #
 (
  parameter WIDTH = 8
 )
 (
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED
 );

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end
  end

  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      if(count == 1023) begin
        LED <= LED + 1;
      end 
    end
  end

endmodule
"""

def test():
    veriloggen.reset()
    test_module = stub_withcode.mkTop()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import copy

import veriloggen.core.vtypes as vtypes
from veriloggen.core.vtypes import _Numeric
from veriloggen.core.vtypes import _Variable
from veriloggen.core.vtypes import _Constant


def to_fixed(value, point, signed=False):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return value

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return int(value * mag)

    if isinstance(value, (int, bool)):
        mag = vtypes.Int(2) ** point
        return vtypes.Int(value) * mag

    if isinstance(value, float):
        mag = vtypes.Int(2) ** point
        return vtypes.Float(value) * mag

    if hasattr(value, 'signed') and value.signed:
        signed = True

    return shift_left(value, point, signed)


def fixed_to_int(value, point, signed=False):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return value

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return int(value / mag)

    if isinstance(value, (int, bool, float)):
        mag = vtypes.Int(2) ** point
        return vtypes.Int(value) / mag

    if hasattr(value, 'signed') and value.signed:
        signed = True

    return shift_right(value, point, signed)


def fixed_to_int_low(value, point):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return 0

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return int(value % mag)

    return vtypes.And(value, vtypes.Repeat(vtypes.Int(1, 1), point))


def fixed_to_real(value, point, signed=False):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return vtypes.SystemTask('itor', value)

    if isinstance(value, float):
        raise TypeError("value is already float.")

    if isinstance(value, (int, bool)) and isinstance(point, int):
        mag = 2 ** point
        return float(value) / mag

    if hasattr(value, 'signed') and value.signed:
        signed = True

    width = value.bit_length()
    msb = (value[width - 1] if isinstance(value, vtypes._Variable) else
           (value >> (width - 1)) & 0x1)

    v0 = (vtypes.SystemTask('itor', fixed_to_int(value, point)) +
          vtypes.SystemTask('itor', fixed_to_int_low(value, point)) /
          vtypes.SystemTask('itor', vtypes.Int(2) ** point))

    nv = vtypes.Unot(value) + 1
    v1 = ((vtypes.SystemTask('itor', fixed_to_int(nv, point)) +
           vtypes.SystemTask('itor', fixed_to_int_low(nv, point)) /
           vtypes.SystemTask('itor', vtypes.Int(2) ** point))) * vtypes.SystemTask('itor', -1)

    return vtypes.Mux(signed and msb == 0, v0, v1)


def adjust(left, right, lpoint, rpoint, signed=True):
    diff_lpoint = vtypes.Mux(rpoint < lpoint, 0, rpoint - lpoint)
    diff_rpoint = vtypes.Mux(lpoint < rpoint, 0, lpoint - rpoint)
    ldata = vtypes.Mux(diff_lpoint == 0, left,
                       shift_left(left, diff_lpoint, signed))
    rdata = vtypes.Mux(diff_rpoint == 0, right,
                       shift_left(right, diff_rpoint, signed))
    _ldata = vtypes.Mux(signed, vtypes.SystemTask('signed', ldata), ldata)
    _rdata = vtypes.Mux(signed, vtypes.SystemTask('signed', rdata), rdata)
    return _ldata, _rdata


def shift_left(value, size, signed=True):
    if isinstance(value, vtypes.Int):
        value = value.value

    if isinstance(value, int) and isinstance(size, int):
        return value << size

    if isinstance(value, bool) and isinstance(size, int):
        return value << size

    return vtypes.Sll(value, size)


def shift_right(value, size, signed=True):
    if isinstance(value, vtypes.Int):
        value = value.value

    if isinstance(value, int) and isinstance(size, int):
        return value >> size

    if isinstance(value, bool) and isinstance(size, int):
        return value >> size

    return vtypes.Mux(signed, vtypes.Sra(value, size), vtypes.Srl(value, size))

#-------------------------------------------------------------------------


def _max_mux(a, b):
    return vtypes.Mux(a > b, a, b)


def _min_mux(a, b):
    return vtypes.Mux(a < b, a, b)

#-------------------------------------------------------------------------


def FixedInput(m, name, width=32, point=0, signed=False):
    var = m.Input(name, width, signed=signed)
    return Fixed(var, point, signed)


def FixedOutput(m, name, width=32, point=0, signed=False):
    var = m.Output(name, width, signed=signed)
    return Fixed(var, point, signed)


def FixedOutputReg(m, name, width=32, point=0, signed=False):
    var = m.OutputReg(name, width, signed=signed)
    return Fixed(var, point, signed)


def FixedReg(m, name, width=32, point=0, signed=False):
    var = m.Reg(name, width, signed=signed)
    return Fixed(var, point, signed)


def FixedWire(m, name, width=32, point=0, signed=False):
    var = m.Wire(name, width, signed=signed)
    return Fixed(var, point, signed)

#-------------------------------------------------------------------------


class Fixed(vtypes.VeriloggenNode):

    def __init__(self, value, point, signed=None, raw=True):
        vtypes.VeriloggenNode.__init__(self)
        self.value = value if raw else to_fixed(value, point)
        self.point = point
        self.signed = vtypes.get_signed(value) if signed is None else signed

    def __hash__(self):
        return hash((id(self), self.object_id))

    def _adjust(self, value):
        lpoint = self.point
        if not isinstance(value, Fixed):
            rvalue = value
            rsigned = vtypes.get_signed(value)
            rpoint = 0
        else:
            rvalue = value.value
            rsigned = value.signed
            rpoint = value.point

        ldiff = vtypes.Mux(lpoint <= rpoint, 0, lpoint - rpoint)
        rdiff = vtypes.Mux(lpoint >= rpoint, 0, rpoint - lpoint)
        v = vtypes.Mux(lpoint > rpoint, shift_left(rvalue, ldiff, rsigned),
                       vtypes.Mux(lpoint < rpoint, shift_right(rvalue, rdiff, rsigned), rvalue))

        return v

    def write(self, value, blk=False, ldelay=None, rdelay=None):
        v = self._adjust(value)
        return self.value.write(v, blk=blk, ldelay=ldelay, rdelay=rdelay)

    def write_raw(self, value, blk=False, ldelay=None, rdelay=None):
        return self.value.write(value, blk=blk, ldelay=ldelay, rdelay=rdelay)

    def read(self):
        return self.value.read()

    def assign(self, value):
        v = self._adjust(value)
        return self.value.assign(v)

    def reset(self):
        return self.value.reset()

    def bit_length(self):
        return self.value.bit_length()

    def get_signed(self):
        return self.signed

    def add(self, r):
        return self.write(self + r)

    def sub(self, r):
        return self.write(self - r)

    def inc(self):
        return self.add(1)

    def dec(self):
        return self.sub(1)

    def __call__(self, value, blk=False, ldelay=None, rdelay=None):
        return self.write(value, blk=blk, ldelay=ldelay, rdelay=rdelay)

    #-------------------------------------------------------------------------
    @property
    def raw(self):
        return self.value

    @property
    def int_part(self):
        return self.value >> self.point

    @property
    def dec_part(self):
        mask = vtypes.Mux(self.point == 0, 0, vtypes.Repeat(
            vtypes.Int(1, width=1), self.point))
        return self.value & mask

    def _binary_op(self, op, r):
        lvalue = self.value
        lpoint = self.point
        lsigned = self.signed

        if not isinstance(r, Fixed):
            rvalue = r
            rsigned = vtypes.get_signed(r)
            rpoint = 0
        else:
            rvalue = r.value
            rsigned = r.signed
            rpoint = r.point

        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned
        ldata, rdata = adjust(lvalue, rvalue, lpoint, rpoint, signed)

        data = op(ldata, rdata)

        return Fixed(data, point, signed)

    def _binary_op_mul(self, op, r):
        lvalue = self.value
        lpoint = self.point
        lsigned = self.signed

        if not isinstance(r, Fixed):
            rvalue = r
            rsigned = vtypes.get_signed(r)
            rpoint = 0
        else:
            rvalue = r.value
            rsigned = r.signed
            rpoint = r.point

        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned
        ldata = vtypes.SystemTask('signed', lvalue)
        rdata = vtypes.SystemTask('signed', rvalue)
        shift_size = _min_mux(lpoint, rpoint)

        data = op(ldata, rdata)
        data = shift_right(data, shift_size, signed=signed)

        return Fixed(data, point, signed)

    def _binary_op_div(self, op, r):
        lvalue = self.value
        lpoint = self.point
        lsigned = self.signed

        if not isinstance(r, Fixed):
            rvalue = r
            rsigned = vtypes.get_signed(r)
            rpoint = 0
        else:
            rvalue = r.value
            rsigned = r.signed
            rpoint = r.point

        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned
        lwidth = lvalue.bit_length()
        rwidth = rvalue.bit_length()
        ldata, rdata = adjust(lvalue, rvalue, lpoint, rpoint, signed)

        try:
            lmsb = ldata[lwidth - 1]
        except:
            lmsb = (ldata >> (lwidth - 1) & 0x1)

        try:
            rmsb = rdata[lwidth - 1]
        except:
            rmsb = (rdata >> (rwidth - 1) & 0x1)

        abs_ldata = (ldata if not lsigned else
                     vtypes.Mux(lmsb == 0, ldata, vtypes.Unot(ldata) + 1))
        abs_rdata = (rdata if not rsigned else
                     vtypes.Mux(rmsb == 0, rdata, vtypes.Unot(rdata) + 1))
        abs_data = op(abs_ldata, abs_rdata)
        data = (abs_data if not signed else
                vtypes.Mux(vtypes.Ors(vtypes.Ands(lmsb, rmsb),
                                      vtypes.Ands(vtypes.Not(lmsb), vtypes.Not(rmsb))),
                           abs_data, vtypes.Unot(abs_data) + 1))

        return Fixed(data, point, signed)

    def _binary_logical_op(self, op, r):
        lvalue = self.value

        if not isinstance(r, Fixed):
            rvalue = r
        else:
            rvalue = r.value

        return op(lvalue, rvalue)

    def __add__(self, r):
        return self._binary_op(vtypes.Plus, r)

    def __sub__(self, r):
        return self._binary_op(vtypes.Minus, r)

    def __mul__(self, r):
        return self._binary_op_mul(vtypes.Times, r)

    def __div__(self, r):
        return self._binary_op_div(vtypes.Divide, r)

    def __truediv__(self, r):
        return self._binary_op_div(vtypes.Divide, r)


def FixedConst(value, point, raw=False):
    return Fixed(value, point, raw=raw)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
import veriloggen.resolver.resolver as resolver

def mkOrigLed():
    m = Module('blinkled')
    width = m.Parameter('WIDTH', 8)
    
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.OutputReg('LED', width)
    
    count = m.Reg('count', 32, width)
    i = m.Integer('i')

    m.Always(Posedge(clk))(
        If(rst)(
            For(i(0), i<width, i.inc())(
                count[i](0)
            )
        ).Else(
            For(i(0), i<width, i.inc())(
                If(count[i] == 1023)(
                    count[i](0)
                ).Else(
                    count[i](count[i] + i)
                )
            )
        ))
    
    _sum = m.Reg('_sum', 32)

    m.Always(Posedge(clk))(
        If(rst)(
            led( 0 )
        ).Else(
            _sum(0, blk=True),
            For(i(0), i<width, i.inc())(
                _sum(_sum + count[i], blk=True)
            ),
            If(_sum == Int(1024 - 1) * width)(
                led(led + 1)
            )
        ))

    return m

def mkLed():
    led = mkOrigLed()
    # resolve() is idempotent
    led = resolver.resolve(led)
    led = resolver.resolve(led)
    led = resolver.resolve(led)
    return resolver.resolve(led)

if __name__ == '__main__':
    led = mkLed()
    verilog = led.to_verilog()
    print(verilog)



from setuptools import setup, find_packages

import utils.version
import re
import os

m = re.search(r'(\d+\.\d+\.\d+(-.+)?)', utils.version.VERSION)
version = m.group(1) if m is not None else '0.0.0'

def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(name='veriloggen',
      version=version,
      description='A library for constructing a Verilog HDL source code in Python',
      long_description=read('README.rst'),
      keywords = 'FPGA, Verilog HDL',
      author='Shinya Takamaeda-Yamazaki',
      #author_email='shinya.takamaeda_at_gmail_com',
      license="Apache License 2.0",
      url='https://github.com/PyHDI/veriloggen',
      packages=find_packages(),
      #package_data={ 'path' : ['*.*'], },
      install_requires=[ 'pyverilog>=1.0.6', 'Jinja2>=2.8' ],
      extras_require={
          'graph' : [ 'pygraphviz>=1.3.1' ],
          'test' : [ 'pytest>=2.8.2', 'pytest-pythonpath>=0.7' ],
      },
)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import _list

expected_verilog = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED_0,
  output reg [WIDTH-1:0] LED_1,
  output reg [WIDTH-1:0] LED_2,
  output reg [WIDTH-1:0] LED_3,
  output reg [WIDTH-1:0] LED_4,
  output reg [WIDTH-1:0] LED_5,
  output reg [WIDTH-1:0] LED_6,
  output reg [WIDTH-1:0] LED_7
);

  reg [32-1:0] count;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end
  end


  always @(*) begin
    case(count % 8)
      0: begin
        LED_0 = count[WIDTH-1:0];
      end
      1: begin
        LED_1 = count[WIDTH-1:0];
      end
      2: begin
        LED_2 = count[WIDTH-1:0];
      end
      3: begin
        LED_3 = count[WIDTH-1:0];
      end
      4: begin
        LED_4 = count[WIDTH-1:0];
      end
      5: begin
        LED_5 = count[WIDTH-1:0];
      end
      6: begin
        LED_6 = count[WIDTH-1:0];
      end
      7: begin
        LED_7 = count[WIDTH-1:0];
      end
    endcase
  end
endmodule
"""

def test():
    veriloggen.reset()
    test_module = _list.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import os
import sys

import veriloggen.core.vtypes as vtypes
from veriloggen.fsm.fsm import FSM

from .thread import Thread


class ThreadPool(vtypes.VeriloggenNode):
    __intrinsics__ = ('run', 'join', 'done', 'reset', 'ret')

    def __init__(self, m, name, clk, rst, targ, numthreads, datawidth=32):
        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.targ = targ
        self.numthreads = numthreads
        self.datawidth = datawidth
        self.threads = [Thread(m, '_'.join([name, str(i)]), clk, rst, targ,
                               datawidth=datawidth, tid=i)
                        for i in range(numthreads)]

        self.start = self.m.Reg(
            '_'.join(['', self.name, 'start']), self.numthreads, initval=0)

    def start(self, tid, *args, **kwargs):
        return self.threads[tid].start(*args, **kwargs)

    def extend(self, tid, fsm, *args, **kwargs):
        return self.threads[tid].extend(fsm, *args, **kwargs)

    def run(self, fsm, tid, *args, **kwargs):
        """ start as a child thread """

        fsm(
            self.start[tid](1)
        )
        fsm.goto_next()

        for thread in self.threads:
            start_cond = self.start[thread.tid]

            if not thread.is_child and thread.end_state is not None:
                raise ValueError('thread %d already started' % thread.tid)

            if thread.fsm is None:
                thread.fsm = FSM(thread.m, thread.name, thread.clk, thread.rst)

            thread.is_child = True

            if thread.start_state is not None:
                thread.fsm._set_index(thread.start_state)
            else:
                thread.start_state = thread.fsm.current

        fsm_start = fsm.current

        for thread in self.threads:
            start_cond = self.start[thread.tid]

            fsm._set_index(fsm_start)

            thread._synthesize_run_fsm(fsm, args, kwargs, cond=start_cond)

            if thread.end_state is None:
                thread.end_state = thread.fsm.current

        fsm(
            self.start[tid](0)
        )
        fsm.goto_next()

        return 0

    def join(self, fsm, tid):
        """ wait for the completion """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            end_flag = (thread.fsm.state == thread.end_state)
            patterns.append(((tid == thread.tid), end_flag))

        patterns.append((None, 0))
        cond = vtypes.PatternMux(*patterns)
        fsm.If(cond).goto_next()

        return 0

    def done(self, fsm, tid):
        """ check whethe the thread is running """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            end_flag = (thread.fsm.state == thread.end_state)
            patterns.append(((tid == thread.tid), end_flag))

        patterns.append((None, 0))
        cond = vtypes.PatternMux(*patterns)

        return cond

    def reset(self, fsm, tid):
        """ reset the FSM counter to the initial state """

        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)

        for thread in self.threads:
            reset_flag = vtypes.Land(
                (fsm.state == fsm.current), (tid == thread.tid))

            thread.fsm._set_index(thread.end_state)

            if thread.called is not None:
                thread.fsm.If(reset_flag)(
                    thread.called(0)
                )

            thread.fsm.goto_from(
                thread.end_state, thread.start_state, reset_flag)
            thread._set_index(thread.start_state)

        return 0

    def ret(self, fsm, tid):
        """ return value """

        patterns = []
        for thread in self.threads:
            if thread.end_state is None:
                raise ValueError('thread %d not started' % thread.tid)
            patterns.append(((tid == thread.tid), thread.return_value))

        patterns.append((None, vtypes.IntX()))
        ret = vtypes.PatternMux(*patterns)

        return ret



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import dataflow_matmul_bram

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg [8-1:0] bram_addr0;
  reg [32-1:0] bram_din0;
  reg bram_we0;
  wire [32-1:0] bram_dout0;
  reg [8-1:0] bram_addr1;
  reg [32-1:0] bram_din1;
  reg bram_we1;
  wire [32-1:0] bram_dout1;
  reg [8-1:0] bram_addr2;
  reg [32-1:0] bram_din2;
  reg bram_we2;
  wire [32-1:0] bram_dout2;
  reg start;
  wire busy;

  Matmul
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .bram_addr0(bram_addr0),
    .bram_din0(bram_din0),
    .bram_we0(bram_we0),
    .bram_dout0(bram_dout0),
    .bram_addr1(bram_addr1),
    .bram_din1(bram_din1),
    .bram_we1(bram_we1),
    .bram_dout1(bram_dout1),
    .bram_addr2(bram_addr2),
    .bram_din2(bram_din2),
    .bram_we2(bram_we2),
    .bram_dout2(bram_dout2),
    .start(start),
    .busy(busy)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    start = 0;
    bram_addr0 = 0;
    bram_din0 = 0;
    bram_we0 = 0;
    bram_addr1 = 0;
    bram_din1 = 0;
    bram_we1 = 0;
    bram_addr2 = 0;
    bram_din2 = 0;
    bram_we2 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #100000;
    $finish;
  end

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _d1_fsm;
  reg _fsm_cond_4_0_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _d1_fsm <= fsm_init;
      _fsm_cond_4_0_1 <= 0;
    end else begin
      _d1_fsm <= fsm;
      case(_d1_fsm)
        fsm_4: begin
          if(_fsm_cond_4_0_1) begin
            start <= 0;
          end 
        end
      endcase
      case(fsm)
        fsm_init: begin
          if(reset_done) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          bram_addr0 <= -1;
          bram_addr1 <= -1;
          bram_addr2 <= -1;
          fsm <= fsm_2;
        end
        fsm_2: begin
          bram_addr0 <= bram_addr0 + 1;
          bram_din0 <= (bram_addr0 + 1) % 256;
          bram_we0 <= 1;
          if(bram_we0 && (bram_addr0 == 255)) begin
            bram_we0 <= 0;
          end 
          bram_addr1 <= bram_addr1 + 1;
          bram_din1 <= ((bram_addr1 + 1) % 256 % 16 == (bram_addr1 + 1) % 256 / 16)? 2 : 0;
          bram_we1 <= 1;
          if(bram_we1 && (bram_addr1 == 255)) begin
            bram_we1 <= 0;
          end 
          bram_addr2 <= bram_addr2 + 1;
          bram_din2 <= ((bram_addr2 + 1) % 256 % 16 == (bram_addr2 + 1) % 256 / 16)? 2 : 0;
          bram_we2 <= 1;
          if(bram_we2 && (bram_addr2 == 255)) begin
            bram_we2 <= 0;
          end 
          if(bram_we2 && (bram_addr0 == 255)) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(!busy) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          start <= 1;
          _fsm_cond_4_0_1 <= 1;
          fsm <= fsm_5;
        end
        fsm_5: begin
          if(busy) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if(!busy) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          $finish;
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(uut.zwe) begin
      $display("zaddr=%10d zdout=%10d", uut.zaddr, uut.zdout);
      if(uut.zaddr * 2 != uut.zdout) begin
        $display("## wrong result");
      end 
    end 
  end


endmodule



module Matmul
(
  input CLK,
  input RST,
  input [8-1:0] bram_addr0,
  input [32-1:0] bram_din0,
  input bram_we0,
  output [32-1:0] bram_dout0,
  input [8-1:0] bram_addr1,
  input [32-1:0] bram_din1,
  input bram_we1,
  output [32-1:0] bram_dout1,
  input [8-1:0] bram_addr2,
  input [32-1:0] bram_din2,
  input bram_we2,
  output [32-1:0] bram_dout2,
  input start,
  output reg busy
);

  wire [8-1:0] _addr0;
  wire [32-1:0] _din0;
  wire _we0;
  wire [32-1:0] _dout0;
  wire [8-1:0] _addr1;
  wire [32-1:0] _din1;
  wire _we1;
  wire [32-1:0] _dout1;
  wire [8-1:0] _addr2;
  wire [32-1:0] _din2;
  wire _we2;
  wire [32-1:0] _dout2;

  BRAM2
  inst_bram0
  (
    .CLK(CLK),
    .ADDR0(bram_addr0),
    .DIN0(bram_din0),
    .WE0(bram_we0),
    .DOUT0(bram_dout0),
    .ADDR1(_addr0),
    .DIN1(_din0),
    .WE1(_we0),
    .DOUT1(_dout0)
  );


  BRAM2
  inst_bram1
  (
    .CLK(CLK),
    .ADDR0(bram_addr1),
    .DIN0(bram_din1),
    .WE0(bram_we1),
    .DOUT0(bram_dout1),
    .ADDR1(_addr1),
    .DIN1(_din1),
    .WE1(_we1),
    .DOUT1(_dout1)
  );


  BRAM2
  inst_bram2
  (
    .CLK(CLK),
    .ADDR0(bram_addr2),
    .DIN0(bram_din2),
    .WE0(bram_we2),
    .DOUT0(bram_dout2),
    .ADDR1(_addr2),
    .DIN1(_din2),
    .WE1(_we2),
    .DOUT1(_dout2)
  );

  reg [8-1:0] xaddr;
  wire [32-1:0] xdin;
  reg [8-1:0] yaddr;
  wire [32-1:0] ydin;
  reg [8-1:0] zaddr;
  reg [32-1:0] zdout;
  reg zwe;
  assign _addr0 = xaddr;
  assign _we0 = 0;
  assign xdin = _dout0;
  assign _addr1 = yaddr;
  assign _we1 = 0;
  assign ydin = _dout1;
  assign _addr2 = zaddr;
  assign _din2 = zdout;
  assign _we2 = zwe;
  reg ivalid;
  wire [32-1:0] odata;
  wire ovalid;
  reg vreset;

  madd
  madd
  (
    .CLK(CLK),
    .RST(RST),
    .xd(xdin),
    .xv(ivalid),
    .yd(ydin),
    .yv(ivalid),
    .zd(odata),
    .zv(ovalid),
    .vreset_data(vreset),
    .vreset(ivalid)
  );

  reg [5-1:0] _tmp_0;
  reg [32-1:0] _tmp_1;
  reg [5-1:0] _tmp_2;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] _d1_fsm;
  reg _fsm_cond_1_0_1;
  reg _fsm_cond_1_1_1;
  reg _fsm_cond_1_2_1;
  reg _fsm_cond_1_3_1;
  reg _fsm_cond_2_4_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _d1_fsm <= fsm_init;
      xaddr <= 0;
      yaddr <= 0;
      zaddr <= 0;
      busy <= 0;
      vreset <= 0;
      ivalid <= 0;
      _tmp_0 <= 0;
      _tmp_1 <= 0;
      _tmp_2 <= 0;
      _fsm_cond_1_0_1 <= 0;
      _fsm_cond_1_1_1 <= 0;
      _fsm_cond_1_2_1 <= 0;
      _fsm_cond_1_3_1 <= 0;
      zdout <= 0;
      _fsm_cond_2_4_1 <= 0;
    end else begin
      _d1_fsm <= fsm;
      case(_d1_fsm)
        fsm_1: begin
          if(_fsm_cond_1_0_1) begin
            vreset <= 0;
          end 
          if(_fsm_cond_1_1_1) begin
            vreset <= 1;
          end 
          if(_fsm_cond_1_2_1) begin
            ivalid <= 0;
          end 
          if(_fsm_cond_1_3_1) begin
            ivalid <= 1;
          end 
        end
        fsm_2: begin
          if(_fsm_cond_2_4_1) begin
            zwe <= 0;
          end 
        end
      endcase
      case(fsm)
        fsm_init: begin
          xaddr <= -1;
          yaddr <= -1;
          zaddr <= -1;
          zwe <= 0;
          busy <= 0;
          vreset <= 0;
          ivalid <= 0;
          _tmp_0 <= 0;
          _tmp_1 <= 0;
          _tmp_2 <= 0;
          if(start) begin
            busy <= 1;
          end 
          if(start) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          _fsm_cond_1_0_1 <= 1;
          _fsm_cond_1_1_1 <= _tmp_0 == 0;
          if(_tmp_0 < 16) begin
            xaddr <= xaddr + 1;
            yaddr <= yaddr + 1;
          end 
          _fsm_cond_1_2_1 <= 1;
          _fsm_cond_1_3_1 <= _tmp_0 < 16;
          if(_tmp_0 < 16) begin
            _tmp_0 <= _tmp_0 + 1;
          end 
          if(ovalid) begin
            _tmp_2 <= _tmp_2 + 1;
          end 
          if(ovalid) begin
            _tmp_1 <= odata;
          end 
          if(_tmp_2 == 16) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          zaddr <= zaddr + 1;
          zdout <= _tmp_1;
          zwe <= 1;
          _fsm_cond_2_4_1 <= 1;
          if(yaddr == 255) begin
            yaddr <= -1;
          end 
          if(yaddr < 255) begin
            xaddr <= xaddr - 16;
          end 
          _tmp_0 <= 0;
          _tmp_2 <= 0;
          if(!(zaddr == 254)) begin
            fsm <= fsm_1;
          end 
          if(zaddr == 254) begin
            fsm <= fsm_init;
          end 
        end
      endcase
    end
  end


endmodule



module BRAM2
(
  input CLK,
  input [8-1:0] ADDR0,
  input [32-1:0] DIN0,
  input WE0,
  output [32-1:0] DOUT0,
  input [8-1:0] ADDR1,
  input [32-1:0] DIN1,
  input WE1,
  output [32-1:0] DOUT1
);

  reg [8-1:0] delay_ADDR0;
  reg [8-1:0] delay_ADDR1;
  reg [32-1:0] mem [0:256-1];

  always @(posedge CLK) begin
    if(WE0) begin
      mem[ADDR0] <= DIN0;
    end 
    delay_ADDR0 <= ADDR0;
  end

  assign DOUT0 = mem[delay_ADDR0];

  always @(posedge CLK) begin
    if(WE1) begin
      mem[ADDR1] <= DIN1;
    end 
    delay_ADDR1 <= ADDR1;
  end

  assign DOUT1 = mem[delay_ADDR1];

endmodule



module madd
(
  input CLK,
  input RST,
  input signed [32-1:0] xd,
  input xv,
  input signed [32-1:0] yd,
  input yv,
  input [1-1:0] vreset_data,
  input vreset,
  output signed [32-1:0] zd,
  output zv
);

  wire signed [32-1:0] _tmp_data_0;
  wire _tmp_valid_0;
  wire _tmp_ready_0;
  wire signed [64-1:0] _tmp_odata_0;
  reg signed [64-1:0] _tmp_data_reg_0;
  assign _tmp_data_0 = _tmp_data_reg_0;
  wire _tmp_ovalid_0;
  reg _tmp_valid_reg_0;
  assign _tmp_valid_0 = _tmp_valid_reg_0;
  wire _tmp_enable_0;
  wire _tmp_update_0;
  assign _tmp_enable_0 = (_tmp_ready_0 || !_tmp_valid_0) && 1 && (xv && yv);
  assign _tmp_update_0 = _tmp_ready_0 || !_tmp_valid_0;

  multiplier_0
  mul0
  (
    .CLK(CLK),
    .RST(RST),
    .update(_tmp_update_0),
    .enable(_tmp_enable_0),
    .valid(_tmp_ovalid_0),
    .a(xd),
    .b(yd),
    .c(_tmp_odata_0)
  );

  assign _tmp_ready_0 = (_tmp_ready_8 || !_tmp_valid_8) && (_tmp_valid_0 && _tmp_valid_7);
  reg [1-1:0] _tmp_data_1;
  reg _tmp_valid_1;
  wire _tmp_ready_1;
  assign _tmp_ready_1 = (_tmp_ready_2 || !_tmp_valid_2) && _tmp_valid_1;
  reg [1-1:0] _tmp_data_2;
  reg _tmp_valid_2;
  wire _tmp_ready_2;
  assign _tmp_ready_2 = (_tmp_ready_3 || !_tmp_valid_3) && _tmp_valid_2;
  reg [1-1:0] _tmp_data_3;
  reg _tmp_valid_3;
  wire _tmp_ready_3;
  assign _tmp_ready_3 = (_tmp_ready_4 || !_tmp_valid_4) && _tmp_valid_3;
  reg [1-1:0] _tmp_data_4;
  reg _tmp_valid_4;
  wire _tmp_ready_4;
  assign _tmp_ready_4 = (_tmp_ready_5 || !_tmp_valid_5) && _tmp_valid_4;
  reg [1-1:0] _tmp_data_5;
  reg _tmp_valid_5;
  wire _tmp_ready_5;
  assign _tmp_ready_5 = (_tmp_ready_6 || !_tmp_valid_6) && _tmp_valid_5;
  reg [1-1:0] _tmp_data_6;
  reg _tmp_valid_6;
  wire _tmp_ready_6;
  assign _tmp_ready_6 = (_tmp_ready_7 || !_tmp_valid_7) && _tmp_valid_6;
  reg [1-1:0] _tmp_data_7;
  reg _tmp_valid_7;
  wire _tmp_ready_7;
  assign _tmp_ready_7 = (_tmp_ready_8 || !_tmp_valid_8) && (_tmp_valid_0 && _tmp_valid_7);
  reg signed [32-1:0] _tmp_data_8;
  reg _tmp_valid_8;
  wire _tmp_ready_8;
  assign zd = _tmp_data_8;
  assign zv = _tmp_valid_8;
  assign _tmp_ready_8 = 1;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_reg_0 <= 0;
      _tmp_valid_reg_0 <= 0;
      _tmp_data_1 <= 0;
      _tmp_valid_1 <= 0;
      _tmp_data_2 <= 0;
      _tmp_valid_2 <= 0;
      _tmp_data_3 <= 0;
      _tmp_valid_3 <= 0;
      _tmp_data_4 <= 0;
      _tmp_valid_4 <= 0;
      _tmp_data_5 <= 0;
      _tmp_valid_5 <= 0;
      _tmp_data_6 <= 0;
      _tmp_valid_6 <= 0;
      _tmp_data_7 <= 0;
      _tmp_valid_7 <= 0;
      _tmp_data_8 <= 1'd0;
      _tmp_valid_8 <= 0;
    end else begin
      if(_tmp_ready_0 || !_tmp_valid_0) begin
        _tmp_data_reg_0 <= _tmp_odata_0;
      end 
      if(_tmp_ready_0 || !_tmp_valid_0) begin
        _tmp_valid_reg_0 <= _tmp_ovalid_0;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1 && vreset) begin
        _tmp_data_1 <= vreset_data;
      end 
      if(_tmp_valid_1 && _tmp_ready_1) begin
        _tmp_valid_1 <= 0;
      end 
      if((_tmp_ready_1 || !_tmp_valid_1) && 1) begin
        _tmp_valid_1 <= vreset;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && _tmp_ready_1 && _tmp_valid_1) begin
        _tmp_data_2 <= _tmp_data_1;
      end 
      if(_tmp_valid_2 && _tmp_ready_2) begin
        _tmp_valid_2 <= 0;
      end 
      if((_tmp_ready_2 || !_tmp_valid_2) && _tmp_ready_1) begin
        _tmp_valid_2 <= _tmp_valid_1;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && _tmp_ready_2 && _tmp_valid_2) begin
        _tmp_data_3 <= _tmp_data_2;
      end 
      if(_tmp_valid_3 && _tmp_ready_3) begin
        _tmp_valid_3 <= 0;
      end 
      if((_tmp_ready_3 || !_tmp_valid_3) && _tmp_ready_2) begin
        _tmp_valid_3 <= _tmp_valid_2;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && _tmp_ready_3 && _tmp_valid_3) begin
        _tmp_data_4 <= _tmp_data_3;
      end 
      if(_tmp_valid_4 && _tmp_ready_4) begin
        _tmp_valid_4 <= 0;
      end 
      if((_tmp_ready_4 || !_tmp_valid_4) && _tmp_ready_3) begin
        _tmp_valid_4 <= _tmp_valid_3;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && _tmp_ready_4 && _tmp_valid_4) begin
        _tmp_data_5 <= _tmp_data_4;
      end 
      if(_tmp_valid_5 && _tmp_ready_5) begin
        _tmp_valid_5 <= 0;
      end 
      if((_tmp_ready_5 || !_tmp_valid_5) && _tmp_ready_4) begin
        _tmp_valid_5 <= _tmp_valid_4;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && _tmp_ready_5 && _tmp_valid_5) begin
        _tmp_data_6 <= _tmp_data_5;
      end 
      if(_tmp_valid_6 && _tmp_ready_6) begin
        _tmp_valid_6 <= 0;
      end 
      if((_tmp_ready_6 || !_tmp_valid_6) && _tmp_ready_5) begin
        _tmp_valid_6 <= _tmp_valid_5;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_6 && _tmp_valid_6) begin
        _tmp_data_7 <= _tmp_data_6;
      end 
      if(_tmp_valid_7 && _tmp_ready_7) begin
        _tmp_valid_7 <= 0;
      end 
      if((_tmp_ready_7 || !_tmp_valid_7) && _tmp_ready_6) begin
        _tmp_valid_7 <= _tmp_valid_6;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && (_tmp_ready_0 && _tmp_ready_7) && (_tmp_valid_0 && _tmp_valid_7)) begin
        _tmp_data_8 <= _tmp_data_8 + _tmp_data_0;
      end 
      if(_tmp_valid_8 && _tmp_ready_8) begin
        _tmp_valid_8 <= 0;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && (_tmp_ready_0 && _tmp_ready_7)) begin
        _tmp_valid_8 <= _tmp_valid_0 && _tmp_valid_7;
      end 
      if((_tmp_ready_8 || !_tmp_valid_8) && (_tmp_ready_0 && _tmp_ready_7) && (_tmp_valid_0 && _tmp_valid_7) && _tmp_data_7) begin
        _tmp_data_8 <= 1'd0 + _tmp_data_0;
      end 
    end
  end


endmodule



module multiplier_0
(
  input CLK,
  input RST,
  input update,
  input enable,
  output valid,
  input [32-1:0] a,
  input [32-1:0] b,
  output [64-1:0] c
);

  reg valid_reg0;
  reg valid_reg1;
  reg valid_reg2;
  reg valid_reg3;
  reg valid_reg4;
  reg valid_reg5;
  assign valid = valid_reg5;

  always @(posedge CLK) begin
    if(RST) begin
      valid_reg0 <= 0;
      valid_reg1 <= 0;
      valid_reg2 <= 0;
      valid_reg3 <= 0;
      valid_reg4 <= 0;
      valid_reg5 <= 0;
    end else begin
      if(update) begin
        valid_reg0 <= enable;
        valid_reg1 <= valid_reg0;
        valid_reg2 <= valid_reg1;
        valid_reg3 <= valid_reg2;
        valid_reg4 <= valid_reg3;
        valid_reg5 <= valid_reg4;
      end 
    end
  end


  multiplier_core_0
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_0
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [32-1:0] b,
  output [64-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [32-1:0] _b;
  reg signed [64-1:0] _tmpval0;
  reg signed [64-1:0] _tmpval1;
  reg signed [64-1:0] _tmpval2;
  reg signed [64-1:0] _tmpval3;
  reg signed [64-1:0] _tmpval4;
  wire signed [64-1:0] rslt;
  assign rslt = _a * _b;
  assign c = _tmpval4;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _tmpval0 <= rslt;
      _tmpval1 <= _tmpval0;
      _tmpval2 <= _tmpval1;
      _tmpval3 <= _tmpval2;
      _tmpval4 <= _tmpval3;
    end 
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = dataflow_matmul_bram.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import axi_vecadd

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg [32-1:0] slave_awaddr;
  reg [8-1:0] slave_awlen;
  reg slave_awvalid;
  wire slave_awready;
  reg [32-1:0] slave_wdata;
  reg [4-1:0] slave_wstrb;
  reg slave_wlast;
  reg slave_wvalid;
  wire slave_wready;
  reg [32-1:0] slave_araddr;
  reg [8-1:0] slave_arlen;
  reg slave_arvalid;
  wire slave_arready;
  wire [32-1:0] slave_rdata;
  wire slave_rlast;
  wire slave_rvalid;
  reg slave_rready;
  wire [32-1:0] master_awaddr;
  wire [8-1:0] master_awlen;
  wire master_awvalid;
  reg master_awready;
  wire [32-1:0] master_wdata;
  wire [4-1:0] master_wstrb;
  wire master_wlast;
  wire master_wvalid;
  reg master_wready;
  wire [32-1:0] master_araddr;
  wire [8-1:0] master_arlen;
  wire master_arvalid;
  reg master_arready;
  reg [32-1:0] master_rdata;
  reg master_rlast;
  reg master_rvalid;
  wire master_rready;
  reg [32-1:0] waddr;
  localparam waddr_init = 0;
  reg [32-1:0] _awlen;
  reg [32-1:0] raddr;
  localparam raddr_init = 0;
  reg [32-1:0] _arlen;
  reg [32-1:0] _d1_raddr;
  reg _raddr_cond_3_0_1;
  wire _tmp_0;
  assign _tmp_0 = 1;

  always @(*) begin
    slave_awvalid <= _tmp_0;
  end

  wire [32-1:0] _tmp_1;
  assign _tmp_1 = 256;

  always @(*) begin
    slave_awaddr <= _tmp_1;
  end

  wire [8-1:0] _tmp_2;
  assign _tmp_2 = 0;

  always @(*) begin
    slave_awlen <= _tmp_2;
  end

  wire _tmp_3;
  assign _tmp_3 = 1;

  always @(*) begin
    slave_arvalid <= _tmp_3;
  end

  wire [32-1:0] _tmp_4;
  assign _tmp_4 = 256;

  always @(*) begin
    slave_araddr <= _tmp_4;
  end

  wire [8-1:0] _tmp_5;
  assign _tmp_5 = 0;

  always @(*) begin
    slave_arlen <= _tmp_5;
  end

  wire _tmp_6;
  assign _tmp_6 = 1;

  always @(*) begin
    slave_wvalid <= _tmp_6;
  end

  wire [32-1:0] _tmp_7;
  assign _tmp_7 = 512;

  always @(*) begin
    slave_wdata <= _tmp_7;
  end

  wire [4-1:0] _tmp_8;
  assign _tmp_8 = { 4{ 1'd1 } };

  always @(*) begin
    slave_wstrb <= _tmp_8;
  end

  wire _tmp_9;
  assign _tmp_9 = 1;

  always @(*) begin
    slave_wlast <= _tmp_9;
  end

  wire _tmp_10;
  assign _tmp_10 = 1;

  always @(*) begin
    slave_rready <= _tmp_10;
  end


  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .slave_awaddr(slave_awaddr),
    .slave_awlen(slave_awlen),
    .slave_awvalid(slave_awvalid),
    .slave_awready(slave_awready),
    .slave_wdata(slave_wdata),
    .slave_wstrb(slave_wstrb),
    .slave_wlast(slave_wlast),
    .slave_wvalid(slave_wvalid),
    .slave_wready(slave_wready),
    .slave_araddr(slave_araddr),
    .slave_arlen(slave_arlen),
    .slave_arvalid(slave_arvalid),
    .slave_arready(slave_arready),
    .slave_rdata(slave_rdata),
    .slave_rlast(slave_rlast),
    .slave_rvalid(slave_rvalid),
    .slave_rready(slave_rready),
    .master_awaddr(master_awaddr),
    .master_awlen(master_awlen),
    .master_awvalid(master_awvalid),
    .master_awready(master_awready),
    .master_wdata(master_wdata),
    .master_wstrb(master_wstrb),
    .master_wlast(master_wlast),
    .master_wvalid(master_wvalid),
    .master_wready(master_wready),
    .master_araddr(master_araddr),
    .master_arlen(master_arlen),
    .master_arvalid(master_arvalid),
    .master_arready(master_arready),
    .master_rdata(master_rdata),
    .master_rlast(master_rlast),
    .master_rvalid(master_rvalid),
    .master_rready(master_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, slave_awaddr, slave_awlen, slave_awvalid, slave_awready, slave_wdata, slave_wstrb, slave_wlast, slave_wvalid, slave_wready, slave_araddr, slave_arlen, slave_arvalid, slave_arready, slave_rdata, slave_rlast, slave_rvalid, slave_rready, master_awaddr, master_awlen, master_awvalid, master_awready, master_wdata, master_wstrb, master_wlast, master_wvalid, master_wready, master_araddr, master_arlen, master_arvalid, master_arready, master_rdata, master_rlast, master_rvalid, master_rready, waddr, _awlen, raddr, _arlen, _d1_raddr, _raddr_cond_3_0_1, _tmp_0, _tmp_1, _tmp_2, _tmp_3, _tmp_4, _tmp_5, _tmp_6, _tmp_7, _tmp_8, _tmp_9, _tmp_10);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    waddr = waddr_init;
    _awlen = 0;
    raddr = raddr_init;
    _arlen = 0;
    _d1_raddr = raddr_init;
    _raddr_cond_3_0_1 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end

  localparam waddr_1 = 1;
  localparam waddr_2 = 2;
  localparam waddr_3 = 3;
  localparam waddr_4 = 4;
  localparam waddr_5 = 5;
  localparam waddr_6 = 6;
  localparam waddr_7 = 7;
  localparam waddr_8 = 8;

  always @(posedge CLK) begin
    if(RST) begin
      waddr <= waddr_init;
      _awlen <= 0;
    end else begin
      case(waddr)
        waddr_init: begin
          master_awready <= 0;
          master_wready <= 0;
          _awlen <= 0;
          if(master_awvalid) begin
            waddr <= waddr_1;
          end 
        end
        waddr_1: begin
          if(master_awvalid) begin
            master_awready <= 1;
          end 
          waddr <= waddr_2;
        end
        waddr_2: begin
          master_awready <= 0;
          _awlen <= master_awlen;
          waddr <= waddr_3;
        end
        waddr_3: begin
          master_wready <= 0;
          if(master_wvalid) begin
            waddr <= waddr_4;
          end 
        end
        waddr_4: begin
          if(master_wvalid) begin
            master_wready <= 1;
          end 
          waddr <= waddr_5;
        end
        waddr_5: begin
          master_wready <= 0;
          waddr <= waddr_6;
        end
        waddr_6: begin
          waddr <= waddr_7;
        end
        waddr_7: begin
          waddr <= waddr_8;
        end
        waddr_8: begin
          _awlen <= _awlen - 1;
          waddr <= waddr_3;
          if(_awlen == 0) begin
            waddr <= waddr_init;
          end 
        end
      endcase
    end
  end

  localparam raddr_1 = 1;
  localparam raddr_2 = 2;
  localparam raddr_3 = 3;
  localparam raddr_4 = 4;
  localparam raddr_5 = 5;
  localparam raddr_6 = 6;
  localparam raddr_7 = 7;
  localparam raddr_8 = 8;
  localparam raddr_9 = 9;
  localparam raddr_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      raddr <= raddr_init;
      _d1_raddr <= raddr_init;
      _arlen <= 0;
      _raddr_cond_3_0_1 <= 0;
    end else begin
      _d1_raddr <= raddr;
      case(_d1_raddr)
        raddr_3: begin
          if(_raddr_cond_3_0_1) begin
            master_rvalid <= 0;
            master_rlast <= 0;
          end 
        end
      endcase
      case(raddr)
        raddr_init: begin
          master_arready <= 0;
          master_rdata <= -1;
          master_rvalid <= 0;
          master_rlast <= 0;
          if(master_arvalid) begin
            raddr <= raddr_1;
          end 
        end
        raddr_1: begin
          if(master_arvalid) begin
            master_arready <= 1;
            master_rdata <= master_araddr - 1;
          end 
          raddr <= raddr_2;
        end
        raddr_2: begin
          master_arready <= 0;
          _arlen <= master_arlen;
          raddr <= raddr_3;
        end
        raddr_3: begin
          if((master_rready || !master_rvalid) && !master_rlast) begin
            master_rdata <= master_rdata + 1;
            master_rvalid <= 1;
            master_rlast <= 0;
            _arlen <= _arlen - 1;
          end 
          if((master_rready || !master_rvalid) && !master_rlast && (_arlen == 0)) begin
            master_rlast <= 1;
          end 
          _raddr_cond_3_0_1 <= 1;
          raddr <= raddr_4;
        end
        raddr_4: begin
          if(master_rvalid && !master_rready) begin
            master_rvalid <= master_rvalid;
            master_rlast <= master_rlast;
          end 
          if(master_rvalid && master_rready) begin
            raddr <= raddr_5;
          end 
        end
        raddr_5: begin
          raddr <= raddr_6;
        end
        raddr_6: begin
          raddr <= raddr_7;
        end
        raddr_7: begin
          raddr <= raddr_8;
        end
        raddr_8: begin
          if((_arlen + 1 & 255) != 0) begin
            raddr <= raddr_3;
          end 
          if((_arlen + 1 & 255) == 0) begin
            raddr <= raddr_9;
          end 
        end
        raddr_9: begin
          raddr <= raddr_10;
        end
        raddr_10: begin
          raddr <= raddr_init;
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  input [32-1:0] slave_awaddr,
  input [8-1:0] slave_awlen,
  input slave_awvalid,
  output slave_awready,
  input [32-1:0] slave_wdata,
  input [4-1:0] slave_wstrb,
  input slave_wlast,
  input slave_wvalid,
  output slave_wready,
  input [32-1:0] slave_araddr,
  input [8-1:0] slave_arlen,
  input slave_arvalid,
  output slave_arready,
  output reg [32-1:0] slave_rdata,
  output reg slave_rlast,
  output reg slave_rvalid,
  input slave_rready,
  output reg [32-1:0] master_awaddr,
  output reg [8-1:0] master_awlen,
  output reg master_awvalid,
  input master_awready,
  output reg [32-1:0] master_wdata,
  output reg [4-1:0] master_wstrb,
  output reg master_wlast,
  output reg master_wvalid,
  input master_wready,
  output reg [32-1:0] master_araddr,
  output reg [8-1:0] master_arlen,
  output reg master_arvalid,
  input master_arready,
  input [32-1:0] master_rdata,
  input master_rlast,
  input master_rvalid,
  output master_rready
);

  reg [10-1:0] ram_a_0_addr;
  wire [32-1:0] ram_a_0_rdata;
  reg [32-1:0] ram_a_0_wdata;
  reg ram_a_0_wenable;

  ram_a
  inst_ram_a
  (
    .CLK(CLK),
    .ram_a_0_addr(ram_a_0_addr),
    .ram_a_0_rdata(ram_a_0_rdata),
    .ram_a_0_wdata(ram_a_0_wdata),
    .ram_a_0_wenable(ram_a_0_wenable)
  );

  reg [10-1:0] ram_b_0_addr;
  wire [32-1:0] ram_b_0_rdata;
  reg [32-1:0] ram_b_0_wdata;
  reg ram_b_0_wenable;

  ram_b
  inst_ram_b
  (
    .CLK(CLK),
    .ram_b_0_addr(ram_b_0_addr),
    .ram_b_0_rdata(ram_b_0_rdata),
    .ram_b_0_wdata(ram_b_0_wdata),
    .ram_b_0_wenable(ram_b_0_wenable)
  );

  reg [10-1:0] ram_c_0_addr;
  wire [32-1:0] ram_c_0_rdata;
  reg [32-1:0] ram_c_0_wdata;
  reg ram_c_0_wenable;

  ram_c
  inst_ram_c
  (
    .CLK(CLK),
    .ram_c_0_addr(ram_c_0_addr),
    .ram_c_0_rdata(ram_c_0_rdata),
    .ram_c_0_wdata(ram_c_0_wdata),
    .ram_c_0_wenable(ram_c_0_wenable)
  );

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg [32-1:0] _tmp_1;
  reg _tmp_2;
  assign slave_awready = (fsm == 0) && !_tmp_2;
  assign slave_wready = fsm == 1;
  reg [32-1:0] _tmp_fsm_0;
  localparam _tmp_fsm_0_init = 0;
  reg [9-1:0] _tmp_3;
  reg _master_cond_0_1;
  wire _tmp_4;
  wire _tmp_5;
  assign _tmp_4 = 1 && _tmp_ready_8;
  assign _tmp_5 = 1;
  assign master_rready = _tmp_4 && _tmp_5 || _tmp_10 && _tmp_11;
  reg [8-1:0] _tmp_6;
  reg _tmp_7;
  wire [32-1:0] _tmp_data_8;
  wire _tmp_valid_8;
  wire _tmp_ready_8;
  assign _tmp_ready_8 = (_tmp_6 > 0) && !_tmp_7;
  reg _ram_a_cond_0_1;
  reg [32-1:0] _tmp_fsm_1;
  localparam _tmp_fsm_1_init = 0;
  reg [9-1:0] _tmp_9;
  reg _master_cond_1_1;
  wire _tmp_10;
  wire _tmp_11;
  assign _tmp_10 = 1 && _tmp_ready_14;
  assign _tmp_11 = 1;
  reg [8-1:0] _tmp_12;
  reg _tmp_13;
  wire [32-1:0] _tmp_data_14;
  wire _tmp_valid_14;
  wire _tmp_ready_14;
  assign _tmp_ready_14 = (_tmp_12 > 0) && !_tmp_13;
  reg _ram_b_cond_0_1;
  reg _tmp_15;
  reg _tmp_16;
  wire _tmp_17;
  wire _tmp_18;
  assign _tmp_17 = 1 && ((_tmp_ready_61 || !_tmp_valid_61) && (_tmp_15 && _tmp_27));
  assign _tmp_18 = 1;
  localparam _tmp_19 = 1;
  wire [_tmp_19-1:0] _tmp_20;
  assign _tmp_20 = (_tmp_17 || !_tmp_15) && (_tmp_18 || !_tmp_16);
  reg [_tmp_19-1:0] __tmp_20_1;
  wire [32-1:0] _tmp_21;
  reg [32-1:0] __tmp_21_1;
  assign _tmp_21 = (__tmp_20_1)? ram_a_0_rdata : __tmp_21_1;
  reg [8-1:0] _tmp_22;
  reg _tmp_23;
  reg _tmp_24;
  reg _tmp_25;
  reg _tmp_26;
  reg _tmp_27;
  reg _tmp_28;
  wire _tmp_29;
  wire _tmp_30;
  assign _tmp_29 = 1 && ((_tmp_ready_61 || !_tmp_valid_61) && (_tmp_15 && _tmp_27));
  assign _tmp_30 = 1;
  localparam _tmp_31 = 1;
  wire [_tmp_31-1:0] _tmp_32;
  assign _tmp_32 = (_tmp_29 || !_tmp_27) && (_tmp_30 || !_tmp_28);
  reg [_tmp_31-1:0] __tmp_32_1;
  wire [32-1:0] _tmp_33;
  reg [32-1:0] __tmp_33_1;
  assign _tmp_33 = (__tmp_32_1)? ram_b_0_rdata : __tmp_33_1;
  reg [8-1:0] _tmp_34;
  reg _tmp_35;
  reg _tmp_36;
  reg _tmp_37;
  reg _tmp_38;
  reg [8-1:0] _tmp_39;
  reg _tmp_40;
  wire [32-1:0] _tmp_data_41;
  wire _tmp_valid_41;
  wire _tmp_ready_41;
  assign _tmp_ready_41 = (_tmp_39 > 0) && !_tmp_40;
  reg _ram_c_cond_0_1;
  reg [32-1:0] _tmp_fsm_2;
  localparam _tmp_fsm_2_init = 0;
  reg [9-1:0] _tmp_42;
  reg _master_cond_2_1;
  reg _tmp_43;
  reg _tmp_44;
  wire _tmp_45;
  wire _tmp_46;
  assign _tmp_45 = 1 && _tmp_ready_56;
  assign _tmp_46 = 1;
  localparam _tmp_47 = 1;
  wire [_tmp_47-1:0] _tmp_48;
  assign _tmp_48 = (_tmp_45 || !_tmp_43) && (_tmp_46 || !_tmp_44);
  reg [_tmp_47-1:0] __tmp_48_1;
  wire [32-1:0] _tmp_49;
  reg [32-1:0] __tmp_49_1;
  assign _tmp_49 = (__tmp_48_1)? ram_c_0_rdata : __tmp_49_1;
  reg [8-1:0] _tmp_50;
  reg _tmp_51;
  reg _tmp_52;
  reg _tmp_53;
  reg _tmp_54;
  reg _tmp_55;
  wire [32-1:0] _tmp_data_56;
  wire _tmp_valid_56;
  wire _tmp_ready_56;
  assign _tmp_ready_56 = (_tmp_fsm_2 == 3) && ((_tmp_42 > 0) && (master_wready || !master_wvalid));
  reg _master_cond_3_1;
  reg [32-1:0] sum;
  reg _seq_cond_0_1;
  reg [9-1:0] _tmp_57;
  reg [32-1:0] _tmp_58;
  reg _tmp_59;
  assign slave_arready = (fsm == 8) && !_tmp_59;
  reg _tmp_60;
  reg _slave_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_1 <= 0;
      _tmp_0 <= 0;
      _tmp_2 <= 0;
      _tmp_58 <= 0;
      _tmp_57 <= 0;
      _tmp_59 <= 0;
      slave_rdata <= 0;
      slave_rvalid <= 0;
      slave_rlast <= 0;
      _tmp_60 <= 0;
      _slave_cond_0_1 <= 0;
    end else begin
      if(_slave_cond_0_1) begin
        slave_rvalid <= 0;
        slave_rlast <= 0;
        _tmp_60 <= 0;
      end 
      if(slave_awready && slave_awvalid) begin
        _tmp_1 <= slave_awaddr;
        _tmp_0 <= slave_awlen + 1;
      end 
      _tmp_2 <= slave_awready && slave_awvalid;
      if(slave_wready && slave_wvalid && (_tmp_0 > 0)) begin
        _tmp_0 <= _tmp_0 - 1;
      end 
      if(slave_arready && slave_arvalid) begin
        _tmp_58 <= slave_araddr;
        _tmp_57 <= slave_arlen + 1;
      end 
      _tmp_59 <= slave_arready && slave_arvalid;
      if((fsm == 9) && ((_tmp_57 > 0) && (slave_rready || !slave_rvalid) && (_tmp_57 > 0))) begin
        slave_rdata <= sum;
        slave_rvalid <= 1;
        slave_rlast <= 0;
        _tmp_57 <= _tmp_57 - 1;
      end 
      if((fsm == 9) && ((_tmp_57 > 0) && (slave_rready || !slave_rvalid) && (_tmp_57 > 0)) && (_tmp_57 == 1)) begin
        slave_rlast <= 1;
        _tmp_60 <= 1;
      end 
      _slave_cond_0_1 <= 1;
      if(slave_rvalid && !slave_rready) begin
        slave_rvalid <= slave_rvalid;
        slave_rlast <= slave_rlast;
        _tmp_60 <= _tmp_60;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      master_araddr <= 0;
      master_arlen <= 0;
      master_arvalid <= 0;
      _tmp_3 <= 0;
      _master_cond_0_1 <= 0;
      _tmp_9 <= 0;
      _master_cond_1_1 <= 0;
      master_awaddr <= 0;
      master_awlen <= 0;
      master_awvalid <= 0;
      _tmp_42 <= 0;
      _master_cond_2_1 <= 0;
      master_wdata <= 0;
      master_wvalid <= 0;
      master_wlast <= 0;
      master_wstrb <= 0;
      _tmp_55 <= 0;
      _master_cond_3_1 <= 0;
    end else begin
      if(_master_cond_0_1) begin
        master_arvalid <= 0;
      end 
      if(_master_cond_1_1) begin
        master_arvalid <= 0;
      end 
      if(_master_cond_2_1) begin
        master_awvalid <= 0;
      end 
      if(_master_cond_3_1) begin
        master_wvalid <= 0;
        master_wlast <= 0;
        _tmp_55 <= 0;
      end 
      if((_tmp_fsm_0 == 1) && ((master_arready || !master_arvalid) && (_tmp_3 == 0))) begin
        master_araddr <= 1024;
        master_arlen <= 63;
        master_arvalid <= 1;
        _tmp_3 <= 64;
      end 
      _master_cond_0_1 <= 1;
      if(master_arvalid && !master_arready) begin
        master_arvalid <= master_arvalid;
      end 
      if(master_rready && master_rvalid && (_tmp_3 > 0)) begin
        _tmp_3 <= _tmp_3 - 1;
      end 
      if((_tmp_fsm_1 == 1) && ((master_arready || !master_arvalid) && (_tmp_9 == 0))) begin
        master_araddr <= 2048;
        master_arlen <= 63;
        master_arvalid <= 1;
        _tmp_9 <= 64;
      end 
      _master_cond_1_1 <= 1;
      if(master_arvalid && !master_arready) begin
        master_arvalid <= master_arvalid;
      end 
      if(master_rready && master_rvalid && (_tmp_9 > 0)) begin
        _tmp_9 <= _tmp_9 - 1;
      end 
      if((_tmp_fsm_2 == 1) && ((master_awready || !master_awvalid) && (_tmp_42 == 0))) begin
        master_awaddr <= 3072;
        master_awlen <= 63;
        master_awvalid <= 1;
        _tmp_42 <= 64;
      end 
      if((_tmp_fsm_2 == 1) && ((master_awready || !master_awvalid) && (_tmp_42 == 0)) && 0) begin
        master_awvalid <= 0;
      end 
      _master_cond_2_1 <= 1;
      if(master_awvalid && !master_awready) begin
        master_awvalid <= master_awvalid;
      end 
      if(_tmp_valid_56 && ((_tmp_fsm_2 == 3) && ((_tmp_42 > 0) && (master_wready || !master_wvalid))) && ((_tmp_42 > 0) && (master_wready || !master_wvalid) && (_tmp_42 > 0))) begin
        master_wdata <= _tmp_data_56;
        master_wvalid <= 1;
        master_wlast <= 0;
        master_wstrb <= { 4{ 1'd1 } };
        _tmp_42 <= _tmp_42 - 1;
      end 
      if(_tmp_valid_56 && ((_tmp_fsm_2 == 3) && ((_tmp_42 > 0) && (master_wready || !master_wvalid))) && ((_tmp_42 > 0) && (master_wready || !master_wvalid) && (_tmp_42 > 0)) && (_tmp_42 == 1)) begin
        master_wlast <= 1;
        _tmp_55 <= 1;
      end 
      _master_cond_3_1 <= 1;
      if(master_wvalid && !master_wready) begin
        master_wvalid <= master_wvalid;
        master_wlast <= master_wlast;
        _tmp_55 <= _tmp_55;
      end 
    end
  end

  assign _tmp_data_8 = master_rdata;
  assign _tmp_valid_8 = master_rvalid;
  assign _tmp_data_14 = master_rdata;
  assign _tmp_valid_14 = master_rvalid;

  always @(posedge CLK) begin
    if(RST) begin
      ram_a_0_addr <= 0;
      _tmp_6 <= 0;
      ram_a_0_wdata <= 0;
      ram_a_0_wenable <= 0;
      _tmp_7 <= 0;
      _ram_a_cond_0_1 <= 0;
      __tmp_20_1 <= 0;
      __tmp_21_1 <= 0;
      _tmp_26 <= 0;
      _tmp_15 <= 0;
      _tmp_16 <= 0;
      _tmp_24 <= 0;
      _tmp_25 <= 0;
      _tmp_23 <= 0;
      _tmp_22 <= 0;
    end else begin
      if(_ram_a_cond_0_1) begin
        ram_a_0_wenable <= 0;
        _tmp_7 <= 0;
      end 
      if((_tmp_fsm_0 == 2) && (_tmp_6 == 0)) begin
        ram_a_0_addr <= -1;
        _tmp_6 <= 64;
      end 
      if(_tmp_valid_8 && ((_tmp_6 > 0) && !_tmp_7) && (_tmp_6 > 0)) begin
        ram_a_0_addr <= ram_a_0_addr + 1;
        ram_a_0_wdata <= _tmp_data_8;
        ram_a_0_wenable <= 1;
        _tmp_6 <= _tmp_6 - 1;
      end 
      if(_tmp_valid_8 && ((_tmp_6 > 0) && !_tmp_7) && (_tmp_6 == 1)) begin
        _tmp_7 <= 1;
      end 
      _ram_a_cond_0_1 <= 1;
      __tmp_20_1 <= _tmp_20;
      __tmp_21_1 <= _tmp_21;
      if((_tmp_17 || !_tmp_15) && (_tmp_18 || !_tmp_16) && _tmp_24) begin
        _tmp_26 <= 0;
        _tmp_15 <= 0;
        _tmp_16 <= 0;
        _tmp_24 <= 0;
      end 
      if((_tmp_17 || !_tmp_15) && (_tmp_18 || !_tmp_16) && _tmp_23) begin
        _tmp_15 <= 1;
        _tmp_16 <= 1;
        _tmp_26 <= _tmp_25;
        _tmp_25 <= 0;
        _tmp_23 <= 0;
        _tmp_24 <= 1;
      end 
      if((fsm == 4) && (_tmp_22 == 0) && !_tmp_25 && !_tmp_26) begin
        ram_a_0_addr <= 0;
        _tmp_22 <= 63;
        _tmp_23 <= 1;
      end 
      if((_tmp_17 || !_tmp_15) && (_tmp_18 || !_tmp_16) && (_tmp_22 > 0)) begin
        ram_a_0_addr <= ram_a_0_addr + 1;
        _tmp_22 <= _tmp_22 - 1;
        _tmp_23 <= 1;
        _tmp_25 <= 0;
      end 
      if((_tmp_17 || !_tmp_15) && (_tmp_18 || !_tmp_16) && (_tmp_22 == 1)) begin
        _tmp_25 <= 1;
      end 
    end
  end

  reg [32-1:0] _tmp_data_61;
  reg _tmp_valid_61;
  wire _tmp_ready_61;
  assign _tmp_data_41 = _tmp_data_61;
  assign _tmp_valid_41 = _tmp_valid_61;
  assign _tmp_ready_61 = _tmp_ready_41;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_data_61 <= 0;
      _tmp_valid_61 <= 0;
    end else begin
      if((_tmp_ready_61 || !_tmp_valid_61) && (_tmp_17 && _tmp_29) && (_tmp_15 && _tmp_27)) begin
        _tmp_data_61 <= _tmp_21 + _tmp_33;
      end 
      if(_tmp_valid_61 && _tmp_ready_61) begin
        _tmp_valid_61 <= 0;
      end 
      if((_tmp_ready_61 || !_tmp_valid_61) && (_tmp_17 && _tmp_29)) begin
        _tmp_valid_61 <= _tmp_15 && _tmp_27;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      ram_b_0_addr <= 0;
      _tmp_12 <= 0;
      ram_b_0_wdata <= 0;
      ram_b_0_wenable <= 0;
      _tmp_13 <= 0;
      _ram_b_cond_0_1 <= 0;
      __tmp_32_1 <= 0;
      __tmp_33_1 <= 0;
      _tmp_38 <= 0;
      _tmp_27 <= 0;
      _tmp_28 <= 0;
      _tmp_36 <= 0;
      _tmp_37 <= 0;
      _tmp_35 <= 0;
      _tmp_34 <= 0;
    end else begin
      if(_ram_b_cond_0_1) begin
        ram_b_0_wenable <= 0;
        _tmp_13 <= 0;
      end 
      if((_tmp_fsm_1 == 2) && (_tmp_12 == 0)) begin
        ram_b_0_addr <= -1;
        _tmp_12 <= 64;
      end 
      if(_tmp_valid_14 && ((_tmp_12 > 0) && !_tmp_13) && (_tmp_12 > 0)) begin
        ram_b_0_addr <= ram_b_0_addr + 1;
        ram_b_0_wdata <= _tmp_data_14;
        ram_b_0_wenable <= 1;
        _tmp_12 <= _tmp_12 - 1;
      end 
      if(_tmp_valid_14 && ((_tmp_12 > 0) && !_tmp_13) && (_tmp_12 == 1)) begin
        _tmp_13 <= 1;
      end 
      _ram_b_cond_0_1 <= 1;
      __tmp_32_1 <= _tmp_32;
      __tmp_33_1 <= _tmp_33;
      if((_tmp_29 || !_tmp_27) && (_tmp_30 || !_tmp_28) && _tmp_36) begin
        _tmp_38 <= 0;
        _tmp_27 <= 0;
        _tmp_28 <= 0;
        _tmp_36 <= 0;
      end 
      if((_tmp_29 || !_tmp_27) && (_tmp_30 || !_tmp_28) && _tmp_35) begin
        _tmp_27 <= 1;
        _tmp_28 <= 1;
        _tmp_38 <= _tmp_37;
        _tmp_37 <= 0;
        _tmp_35 <= 0;
        _tmp_36 <= 1;
      end 
      if((fsm == 4) && (_tmp_34 == 0) && !_tmp_37 && !_tmp_38) begin
        ram_b_0_addr <= 0;
        _tmp_34 <= 63;
        _tmp_35 <= 1;
      end 
      if((_tmp_29 || !_tmp_27) && (_tmp_30 || !_tmp_28) && (_tmp_34 > 0)) begin
        ram_b_0_addr <= ram_b_0_addr + 1;
        _tmp_34 <= _tmp_34 - 1;
        _tmp_35 <= 1;
        _tmp_37 <= 0;
      end 
      if((_tmp_29 || !_tmp_27) && (_tmp_30 || !_tmp_28) && (_tmp_34 == 1)) begin
        _tmp_37 <= 1;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      ram_c_0_addr <= 0;
      _tmp_39 <= 0;
      ram_c_0_wdata <= 0;
      ram_c_0_wenable <= 0;
      _tmp_40 <= 0;
      _ram_c_cond_0_1 <= 0;
      __tmp_48_1 <= 0;
      __tmp_49_1 <= 0;
      _tmp_54 <= 0;
      _tmp_43 <= 0;
      _tmp_44 <= 0;
      _tmp_52 <= 0;
      _tmp_53 <= 0;
      _tmp_51 <= 0;
      _tmp_50 <= 0;
    end else begin
      if(_ram_c_cond_0_1) begin
        ram_c_0_wenable <= 0;
        _tmp_40 <= 0;
      end 
      if((fsm == 4) && (_tmp_39 == 0)) begin
        ram_c_0_addr <= -1;
        _tmp_39 <= 64;
      end 
      if(_tmp_valid_41 && ((_tmp_39 > 0) && !_tmp_40) && (_tmp_39 > 0)) begin
        ram_c_0_addr <= ram_c_0_addr + 1;
        ram_c_0_wdata <= _tmp_data_41;
        ram_c_0_wenable <= 1;
        _tmp_39 <= _tmp_39 - 1;
      end 
      if(_tmp_valid_41 && ((_tmp_39 > 0) && !_tmp_40) && (_tmp_39 == 1)) begin
        _tmp_40 <= 1;
      end 
      _ram_c_cond_0_1 <= 1;
      __tmp_48_1 <= _tmp_48;
      __tmp_49_1 <= _tmp_49;
      if((_tmp_45 || !_tmp_43) && (_tmp_46 || !_tmp_44) && _tmp_52) begin
        _tmp_54 <= 0;
        _tmp_43 <= 0;
        _tmp_44 <= 0;
        _tmp_52 <= 0;
      end 
      if((_tmp_45 || !_tmp_43) && (_tmp_46 || !_tmp_44) && _tmp_51) begin
        _tmp_43 <= 1;
        _tmp_44 <= 1;
        _tmp_54 <= _tmp_53;
        _tmp_53 <= 0;
        _tmp_51 <= 0;
        _tmp_52 <= 1;
      end 
      if((_tmp_fsm_2 == 2) && (_tmp_50 == 0) && !_tmp_53 && !_tmp_54) begin
        ram_c_0_addr <= 0;
        _tmp_50 <= 63;
        _tmp_51 <= 1;
      end 
      if((_tmp_45 || !_tmp_43) && (_tmp_46 || !_tmp_44) && (_tmp_50 > 0)) begin
        ram_c_0_addr <= ram_c_0_addr + 1;
        _tmp_50 <= _tmp_50 - 1;
        _tmp_51 <= 1;
        _tmp_53 <= 0;
      end 
      if((_tmp_45 || !_tmp_43) && (_tmp_46 || !_tmp_44) && (_tmp_50 == 1)) begin
        _tmp_53 <= 1;
      end 
    end
  end

  assign _tmp_data_56 = _tmp_49;
  assign _tmp_valid_56 = _tmp_43;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;
  localparam fsm_8 = 8;
  localparam fsm_9 = 9;
  localparam fsm_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
    end else begin
      case(fsm)
        fsm_init: begin
          if(_tmp_2) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(slave_wready && slave_wvalid) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          if(_tmp_7) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(_tmp_13) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          fsm <= fsm_5;
        end
        fsm_5: begin
          if(_tmp_40) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if(_tmp_55) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          if(master_wvalid && master_wready && master_wlast) begin
            fsm <= fsm_8;
          end 
        end
        fsm_8: begin
          if(_tmp_59) begin
            fsm <= fsm_9;
          end 
        end
        fsm_9: begin
          if(_tmp_60) begin
            fsm <= fsm_10;
          end 
        end
        fsm_10: begin
          fsm <= fsm_init;
        end
      endcase
    end
  end

  localparam _tmp_fsm_0_1 = 1;
  localparam _tmp_fsm_0_2 = 2;
  localparam _tmp_fsm_0_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_fsm_0 <= _tmp_fsm_0_init;
    end else begin
      case(_tmp_fsm_0)
        _tmp_fsm_0_init: begin
          if(fsm == 2) begin
            _tmp_fsm_0 <= _tmp_fsm_0_1;
          end 
        end
        _tmp_fsm_0_1: begin
          if(master_arready || !master_arvalid) begin
            _tmp_fsm_0 <= _tmp_fsm_0_2;
          end 
        end
        _tmp_fsm_0_2: begin
          _tmp_fsm_0 <= _tmp_fsm_0_3;
        end
        _tmp_fsm_0_3: begin
          if(_tmp_7) begin
            _tmp_fsm_0 <= _tmp_fsm_0_init;
          end 
        end
      endcase
    end
  end

  localparam _tmp_fsm_1_1 = 1;
  localparam _tmp_fsm_1_2 = 2;
  localparam _tmp_fsm_1_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_fsm_1 <= _tmp_fsm_1_init;
    end else begin
      case(_tmp_fsm_1)
        _tmp_fsm_1_init: begin
          if(fsm == 3) begin
            _tmp_fsm_1 <= _tmp_fsm_1_1;
          end 
        end
        _tmp_fsm_1_1: begin
          if(master_arready || !master_arvalid) begin
            _tmp_fsm_1 <= _tmp_fsm_1_2;
          end 
        end
        _tmp_fsm_1_2: begin
          _tmp_fsm_1 <= _tmp_fsm_1_3;
        end
        _tmp_fsm_1_3: begin
          if(_tmp_13) begin
            _tmp_fsm_1 <= _tmp_fsm_1_init;
          end 
        end
      endcase
    end
  end

  localparam _tmp_fsm_2_1 = 1;
  localparam _tmp_fsm_2_2 = 2;
  localparam _tmp_fsm_2_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_fsm_2 <= _tmp_fsm_2_init;
    end else begin
      case(_tmp_fsm_2)
        _tmp_fsm_2_init: begin
          if(fsm == 6) begin
            _tmp_fsm_2 <= _tmp_fsm_2_1;
          end 
        end
        _tmp_fsm_2_1: begin
          if(master_awready || !master_awvalid) begin
            _tmp_fsm_2 <= _tmp_fsm_2_2;
          end 
        end
        _tmp_fsm_2_2: begin
          _tmp_fsm_2 <= _tmp_fsm_2_3;
        end
        _tmp_fsm_2_3: begin
          if(_tmp_55) begin
            _tmp_fsm_2 <= _tmp_fsm_2_init;
          end 
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      sum <= 0;
      _seq_cond_0_1 <= 0;
    end else begin
      if(_seq_cond_0_1) begin
        $display("sum=%d expected_sum=%d", sum, 200640);
      end 
      if(fsm == 0) begin
        sum <= 0;
      end 
      if(master_wvalid && master_wready) begin
        sum <= sum + master_wdata;
      end 
      _seq_cond_0_1 <= master_wvalid && master_wready && master_wlast;
    end
  end


endmodule



module ram_a
(
  input CLK,
  input [10-1:0] ram_a_0_addr,
  output [32-1:0] ram_a_0_rdata,
  input [32-1:0] ram_a_0_wdata,
  input ram_a_0_wenable
);

  reg [10-1:0] ram_a_0_daddr;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_a_0_wenable) begin
      mem[ram_a_0_addr] <= ram_a_0_wdata;
    end 
    ram_a_0_daddr <= ram_a_0_addr;
  end

  assign ram_a_0_rdata = mem[ram_a_0_daddr];

endmodule



module ram_b
(
  input CLK,
  input [10-1:0] ram_b_0_addr,
  output [32-1:0] ram_b_0_rdata,
  input [32-1:0] ram_b_0_wdata,
  input ram_b_0_wenable
);

  reg [10-1:0] ram_b_0_daddr;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_b_0_wenable) begin
      mem[ram_b_0_addr] <= ram_b_0_wdata;
    end 
    ram_b_0_daddr <= ram_b_0_addr;
  end

  assign ram_b_0_rdata = mem[ram_b_0_daddr];

endmodule



module ram_c
(
  input CLK,
  input [10-1:0] ram_c_0_addr,
  output [32-1:0] ram_c_0_rdata,
  input [32-1:0] ram_c_0_wdata,
  input ram_c_0_wenable
);

  reg [10-1:0] ram_c_0_daddr;
  reg [32-1:0] mem [0:1024-1];

  always @(posedge CLK) begin
    if(ram_c_0_wenable) begin
      mem[ram_c_0_addr] <= ram_c_0_wdata;
    end 
    ram_c_0_daddr <= ram_c_0_addr;
  end

  assign ram_c_0_rdata = mem[ram_c_0_daddr];

endmodule
"""


def test():
    veriloggen.reset()
    test_module = axi_vecadd.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)
